# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import os
import unittest

from duplicity.path import *  # pylint: disable=unused-wildcard-import,redefined-builtin
from testing import _runtest_dir
from . import UnitTestCase


class PathTest(UnitTestCase):
    u"""Test basic path functions"""
    def setUp(self):
        super(PathTest, self).setUp()
        self.unpack_testfiles()

    def test_deltree(self):
        u"""Test deleting a tree"""
        assert not os.system(u"cp -pR {0}/testfiles/deltree {0}/testfiles/output".format(_runtest_dir).format(_runtest_dir))
        p = Path(u"{0}/testfiles/output".format(_runtest_dir))
        assert p.isdir()
        p.deltree()
        assert not p.type, p.type

#      def test_compare(self):
#         """Test directory comparisons"""
#         assert not os.system("cp -pR /tmp/testfiles/dir1 /tmp/testfiles/output")
#         assert Path("/tmp/testfiles/dir1").compare_recursive(Path("/tmp/testfiles/output"), 1)
#         assert not Path("/tmp/testfiles/dir1").compare_recursive(Path("/tmp/testfiles/dir2"), 1)

    def test_quote(self):
        u"""Test path quoting"""
        p = Path(u"hello")
        assert p.quote() == u'"hello"'
        assert p.quote(u"\\") == u'"\\\\"', p.quote(u"\\")
        assert p.quote(u"$HELLO") == u'"\\$HELLO"'

    def test_unquote(self):
        u"""Test path unquoting"""
        p = Path(u"foo")  # just to provide unquote function

        def t(s):
            u"""Run test on string s"""
            quoted_version = p.quote(s)
            unquoted = p.unquote(quoted_version)
            assert unquoted == s, (unquoted, s)

        t(u"\\")
        t(u"$HELLO")
        t(u" aoe aoe \\ \n`")

    def test_canonical(self):
        u"""Test getting canonical version of path"""
        c = Path(u".").get_canonical()
        assert c == b".", c

        c = Path(u"//foo/bar/./").get_canonical()
        assert c == b"/foo/bar", c

    def test_compare_verbose(self):
        u"""Run compare_verbose on a few files"""
        vft = Path(u"{0}/testfiles/various_file_types".format(_runtest_dir))
        assert vft.compare_verbose(vft)
        reg_file = vft.append(u"regular_file")
        assert not vft.compare_verbose(reg_file)
        assert reg_file.compare_verbose(reg_file)
        file2 = vft.append(u"executable")
        assert not file2.compare_verbose(reg_file)
        assert file2.compare_verbose(file2)


if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2012 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import gettext
import os
import platform
import subprocess
import sys
import time
import unittest

from duplicity import backend
from duplicity import config
from duplicity import log
from duplicity import util

util.start_debugger()

if sys.version_info.major >= 3:
    gettext.install(u'duplicity', names=[u'ngettext'])
else:
    gettext.install(u'duplicity', names=[u'ngettext'], unicode=True)  # pylint: disable=unexpected-keyword-arg

_testing_dir = os.path.dirname(os.path.abspath(__file__))
_top_dir = os.path.dirname(_testing_dir)
_overrides_dir = os.path.join(_testing_dir, u'overrides')
_bin_dir = os.path.join(_testing_dir, u'overrides', u'bin')

if platform.system().startswith(u'Darwin'):
    # Use temp space from getconf, never /tmp
    _runtest_dir = subprocess.check_output([u'getconf', u'DARWIN_USER_TEMP_DIR'])
    _runtest_dir = util.fsdecode(_runtest_dir).rstrip().rstrip(u'/')
else:
    _runtest_dir = os.getenv(u'TMPDIR', False) or os.getenv(u'TEMP', False) or u'/tmp'

# Adjust python path for duplicity and override modules
sys.path = [_overrides_dir, _top_dir, _bin_dir] + sys.path

# Also set PYTHONPATH for any subprocesses
os.environ[u'PYTHONPATH'] = _overrides_dir + u":" + _top_dir + u":" + os.environ.get(u'PYTHONPATH', u'')

# And PATH for any subprocesses
os.environ[u'PATH'] = _bin_dir + u":" + os.environ.get(u'PATH', u'')

# Now set some variables that help standardize test behavior
os.environ[u'LANG'] = u''
os.environ[u'GNUPGHOME'] = os.path.join(_testing_dir, u'gnupg')

# bzr does not honor perms so fix the perms and avoid annoying error
os.system(u"chmod 700 %s" % os.path.join(_testing_dir, u'gnupg'))

# Standardize time
os.environ[u'TZ'] = u'US/Central'
time.tzset()

# TODO: find place in setup.py to do this
# fix shebangs in _bin_dir to be current python
if sys.version_info.major == 2:
    files = os.listdir(_bin_dir)
    for file in files:
        print(u"converting %s to python2" % file, file=sys.stderr)
        with open(os.path.join(_bin_dir, file), u"r") as f:
            p2 = f.read().replace(u"python3", u"python")
        with open(os.path.join(_bin_dir, file), u"w") as f:
            p2 = f.write(p2)

class DuplicityTestCase(unittest.TestCase):

    sign_key = u'839E6A2856538CCF'
    sign_passphrase = u'test'
    encrypt_key1 = u'839E6A2856538CCF'
    encrypt_key2 = u'453005CE9B736B2A'

    def setUp(self):
        super(DuplicityTestCase, self).setUp()
        self.savedEnviron = {}
        self.savedConfig = {}

        log.setup()
        log.setverbosity(log.WARNING)
        self.set_config(u'print_statistics', 0)
        backend.import_backends()

        self.remove_testfiles()
        self.unpack_testfiles()

        # Have all file references in tests relative to our runtest dir
        os.chdir(_runtest_dir)

    def tearDown(self):
        for key in self.savedEnviron:
            self._update_env(key, self.savedEnviron[key])

        for key in self.savedConfig:
            setattr(config, key, self.savedConfig[key])

        self.remove_testfiles()

        os.chdir(_testing_dir)
        super(DuplicityTestCase, self).tearDown()

    def unpack_testfiles(self):
        assert not os.system(u"rm -rf {0}/testfiles".format(_runtest_dir))
        assert not os.system(u"tar xzf {0}/testfiles.tar.gz -C {1} > /dev/null 2>&1".format(_testing_dir, _runtest_dir))
        assert not os.system(u"mkdir {0}/testfiles/output {0}/testfiles/cache".format(_runtest_dir))

    def remove_testfiles(self):
        assert not os.system(u"rm -rf {0}/testfiles".format(_runtest_dir))

    def _update_env(self, key, value):
        if value is not None:
            os.environ[key] = value
        elif key in os.environ:
            del os.environ[key]

    def set_environ(self, key, value):
        if key not in self.savedEnviron:
            self.savedEnviron[key] = os.environ.get(key)
        self._update_env(key, value)

    def set_config(self, key, value):
        assert hasattr(config, key)
        if key not in self.savedConfig:
            self.savedConfig[key] = getattr(config, key)
        setattr(config, key, value)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; -*-
#
# Copyright 2014 Havard Gulldahl
#
# in part based on dpbxbackend.py:
# Copyright 2013 jno <jno@pisem.net>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# stdlib
import logging
import posixpath

# import duplicity stuff
from duplicity import log
from duplicity.errors import BackendException
import duplicity.backend


def get_jotta_device(jfs):
    jottadev = None
    for j in jfs.devices:  # find Jotta/Shared folder
        if j.name == u'Jotta':
            jottadev = j
    return jottadev


def get_root_dir(jfs):
    jottadev = get_jotta_device(jfs)
    root_dir = jottadev.mountPoints[u'Archive']
    return root_dir


def set_jottalib_logging_level(log_level):
    logger = logging.getLogger(u'jottalib')
    logger.setLevel(getattr(logging, log_level))


def set_jottalib_log_handlers(handlers):
    logger = logging.getLogger(u'jottalib')
    for handler in handlers:
        logger.addHandler(handler)


def get_duplicity_log_level():
    u""" Get the current duplicity log level as a stdlib-compatible logging level"""
    duplicity_log_level = log.LevelName(log.getverbosity())

    # notice is a duplicity-specific logging level not supported by stdlib
    if duplicity_log_level == u'NOTICE':
        duplicity_log_level = u'INFO'

    return duplicity_log_level


class JottaCloudBackend(duplicity.backend.Backend):
    u"""Connect to remote store using JottaCloud API"""

    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        # Import JottaCloud libraries.
        try:
            from jottalib import JFS
            from jottalib.JFS import JFSNotFoundError, JFSIncompleteFile
        except ImportError:
            raise BackendException(u'JottaCloud backend requires jottalib'
                                   u' (see https://pypi.python.org/pypi/jottalib).')

        # Set jottalib loggers to the same verbosity as duplicity
        duplicity_log_level = get_duplicity_log_level()
        set_jottalib_logging_level(duplicity_log_level)

        # Ensure jottalib and duplicity log to the same handlers
        set_jottalib_log_handlers(log._logger.handlers)

        # Will fetch jottacloud auth from environment or .netrc
        self.client = JFS.JFS()

        self.folder = self.get_or_create_directory(parsed_url.path.lstrip(u'/'))
        log.Debug(u"Jottacloud folder for duplicity: %r" % self.folder.path)

    def get_or_create_directory(self, directory_name):
        root_directory = get_root_dir(self.client)
        full_path = posixpath.join(root_directory.path, directory_name)
        try:
            return self.client.getObject(full_path)
        except JFSNotFoundError:
            return root_directory.mkdir(directory_name)

    def _put(self, source_path, remote_filename):
        # - Upload one file
        # - Retried if an exception is thrown
        resp = self.folder.up(source_path.open(), remote_filename)
        log.Debug(u'jottacloud.put(%s,%s): %s' % (source_path.name, remote_filename, resp))

    def _get(self, remote_filename, local_path):
        # - Get one file
        # - Retried if an exception is thrown
        remote_file = self.client.getObject(posixpath.join(self.folder.path, remote_filename))
        log.Debug(u'jottacloud.get(%s,%s): %s' % (remote_filename, local_path.name, remote_file))
        with open(local_path.name, u'wb') as to_file:
            for chunk in remote_file.stream():
                to_file.write(chunk)

    def _list(self):
        # - List all files in the backend
        # - Return a list of filenames
        # - Retried if an exception is thrown
        return list([f.name for f in self.folder.files()  # pylint: disable=no-value-for-parameter
                     if not f.is_deleted() and f.state != u'INCOMPLETE'])

    def _delete(self, filename):
        # - Delete one file
        # - Retried if an exception is thrown
        remote_path = posixpath.join(self.folder.path, filename)
        remote_file = self.client.getObject(remote_path)
        log.Debug(u'jottacloud.delete deleting: %s (%s)' % (remote_file, type(remote_file)))
        remote_file.delete()

    def _query(self, filename):
        u"""Get size of filename"""
        #  - Query metadata of one file
        #  - Return a dict with a 'size' key, and a file size value (-1 for not found)
        #  - Retried if an exception is thrown
        log.Info(u'Querying size of %s' % filename)
        remote_path = posixpath.join(self.folder.path, filename)
        try:
            remote_file = self.client.getObject(remote_path)
        except JFSNotFoundError:
            return {u'size': -1}
        return {
            u'size': remote_file.size,
        }

    def _close(self):
        # - If your backend needs to clean up after itself, do that here.
        pass


duplicity.backend.register_backend(u"jottacloud", JottaCloudBackend)
u""" jottacloud is a Norwegian backup company """



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2013 Matthieu Huin <mhu@enovance.com>
# Copyright 2015 Scott McKenzie <noizyland@gmail.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from builtins import str
import os
import re

import duplicity.backend
from duplicity import config
from duplicity import log
from duplicity.errors import BackendException
from duplicity.util import fsdecode


_VALID_CONTAINER_NAME_RE = re.compile(r"^[a-z0-9](?!.*--)[a-z0-9-]{1,61}[a-z0-9]$")


def _is_valid_container_name(name):
    u"""
    Check, whether the given name conforms to the rules as defined in
    https://docs.microsoft.com/en-us/rest/api/storageservices/naming-and-referencing-containers--blobs--and-metadata
    for valid names.
    """
    match = _VALID_CONTAINER_NAME_RE.match(name)
    return match is not None


class AzureBackend(duplicity.backend.Backend):
    u"""
    Backend for Azure Blob Storage Service
    """
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        # Import Microsoft Azure Storage SDK for Python library.
        try:
            import azure
            import azure.storage
            import azure.storage.blob
            from azure.storage.blob import BlobServiceClient
        except ImportError as e:
            raise BackendException(u"""\
Azure backend requires Microsoft Azure Storage SDK for Python (https://pypi.org/project/azure-storage-blob/).
Exception: %s""" % str(e))

        self.container_name = parsed_url.path.lstrip(u'/')

        if not _is_valid_container_name(self.container_name):
            raise BackendException(u'Invalid Azure Storage Blob container name.')

        if u'AZURE_CONNECTION_STRING' not in os.environ:
            raise BackendException(u'AZURE_CONNECTION_STRING environment variable not set.')

        kwargs = {}

        if config.timeout:
            kwargs[u'timeout'] = config.timeout

        if config.azure_max_single_put_size:
            kwargs[u'max_single_put_size'] = config.azure_max_single_put_size

        if config.azure_max_block_size:
            kwargs[u'max_block_size'] = config.azure_max_single_put_size

        conn_str = os.environ[u'AZURE_CONNECTION_STRING']
        self.blob_service = BlobServiceClient.from_connection_string(conn_str, None, **kwargs)
        self._get_or_create_container()

    def _get_or_create_container(self):
        from azure.core.exceptions import ResourceExistsError

        try:
            self.container = self.blob_service.get_container_client(self.container_name)
            self.container.create_container()
        except ResourceExistsError:
            pass
        except Exception as e:
            log.FatalError(u"Could not create Azure container: %s"
                           % str(e.message).split(u'\n', 1)[0],
                           log.ErrorCode.connection_failed)

    def _put(self, source_path, remote_filename):
        remote_filename = fsdecode(remote_filename)
        kwargs = {}

        if config.azure_max_connections:
            kwargs[u'max_concurrency'] = config.azure_max_connections

        with source_path.open(u"rb") as data:
            self.container.upload_blob(remote_filename, data, **kwargs)

        self._set_tier(remote_filename)

    def _set_tier(self, remote_filename):
        if config.azure_blob_tier is not None:
            self.container.set_standard_blob_tier_blobs(config.azure_blob_tier, remote_filename)

    def _get(self, remote_filename, local_path):
        # https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.containerclient?view=azure-python#download-blob-blob--offset-none--length-none----kwargs-
        blob = self.container.download_blob(remote_filename)
        with local_path.open(u"wb") as download_file:
            download_file.write(blob.readall())

    def _list(self):
        # https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.containerclient?view=azure-python#list-blobs-name-starts-with-none--include-none----kwargs-
        blobs = []
        blob_list = self.container.list_blobs()
        for blob in blob_list:
            blobs.append(blob)

        return [blob.name for blob in blobs]

    def _delete(self, filename):
        # https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.containerclient?view=azure-python#delete-blob-blob--delete-snapshots-none----kwargs-
        self.container.delete_blob(fsdecode(filename))

    def _query(self, filename):
        client = self.container.get_blob_client(fsdecode(filename))
        prop = client.get_blob_properties()
        return {u'size': int(prop.size)}

    def _error_code(self, operation, e):  # pylint: disable=unused-argument
        return log.ErrorCode.backend_not_found


duplicity.backend.register_backend(u'azure', AzureBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from future import standard_library
standard_library.install_aliases()
import os.path
import urllib.request  # pylint: disable=import-error
import urllib.parse  # pylint: disable=import-error
import urllib.error  # pylint: disable=import-error
import re

import duplicity.backend
from duplicity import config
from duplicity import log
from duplicity import tempdir
from duplicity import util


class NCFTPBackend(duplicity.backend.Backend):
    u"""Connect to remote store using File Transfer Protocol"""
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        # we expect an error return, so go low-level and ignore it
        try:
            p = os.popen(u"ncftpls -v")
            fout = p.read()
            ret = p.close()
        except Exception:
            pass
        # the expected error is 8 in the high-byte and some output
        if ret != 0x0800 or not fout:
            log.FatalError(u"NcFTP not found:  Please install NcFTP version 3.1.9 or later",
                           log.ErrorCode.ftp_ncftp_missing)

        # version is the second word of the first line
        version = fout.split(u'\n')[0].split()[1]
        if version < u"3.1.9":
            log.FatalError(u"NcFTP too old:  Duplicity requires NcFTP version 3.1.9,"
                           u"3.2.1 or later.  Version 3.2.0 will not work properly.",
                           log.ErrorCode.ftp_ncftp_too_old)
        elif version == u"3.2.0":
            log.Warn(u"NcFTP (ncftpput) version 3.2.0 may fail with duplicity.\n"
                     u"see: http://www.ncftpd.com/ncftp/doc/changelog.html\n"
                     u"If you have trouble, please upgrade to 3.2.1 or later",
                     log.WarningCode.ftp_ncftp_v320)
        log.Notice(u"NcFTP version is %s" % version)

        self.parsed_url = parsed_url

        self.url_string = duplicity.backend.strip_auth_from_url(self.parsed_url)

        # strip ncftp+ prefix
        self.url_string = duplicity.backend.strip_prefix(self.url_string, u'ncftp')

        # This squelches the "file not found" result from ncftpls when
        # the ftp backend looks for a collection that does not exist.
        # version 3.2.2 has error code 5, 1280 is some legacy value
        self.popen_breaks[u'ncftpls'] = [5, 1280]

        # Use an explicit directory name.
        if self.url_string[-1] != u'/':
            self.url_string += u'/'

        self.password = self.get_password()

        if config.ftp_connection == u'regular':
            self.conn_opt = u'-E'
        else:
            self.conn_opt = u'-F'

        self.tempfd, self.tempname = tempdir.default().mkstemp()
        self.tempfile = os.fdopen(self.tempfd, u"w")
        self.tempfile.write(u"host %s\n" % self.parsed_url.hostname)
        self.tempfile.write(u"user %s\n" % self.parsed_url.username)
        self.tempfile.write(u"pass %s\n" % self.password)
        self.tempfile.close()
        self.flags = u"-f %s %s -t %s -o useCLNT=0,useHELP_SITE=0 " % \
            (self.tempname, self.conn_opt, config.timeout)
        if parsed_url.port is not None and parsed_url.port != 21:
            self.flags += u" -P '%s'" % (parsed_url.port)

    def _put(self, source_path, remote_filename):
        remote_filename = util.fsdecode(remote_filename)
        remote_path = os.path.join(urllib.parse.unquote(re.sub(u'^/', u'', self.parsed_url.path)),
                                   remote_filename).rstrip()
        commandline = u"ncftpput %s -m -V -C '%s' '%s'" % \
            (self.flags, source_path.uc_name, remote_path)
        self.subprocess_popen(commandline)

    def _get(self, remote_filename, local_path):
        remote_filename = util.fsdecode(remote_filename)
        remote_path = os.path.join(urllib.parse.unquote(re.sub(u'^/', u'', self.parsed_url.path)),
                                   remote_filename).rstrip()
        commandline = u"ncftpget %s -V -C '%s' '%s' '%s'" % \
            (self.flags, self.parsed_url.hostname, remote_path.lstrip(u'/'), local_path.uc_name)
        self.subprocess_popen(commandline)

    def _list(self):
        # Do a long listing to avoid connection reset
        commandline = u"ncftpls %s -l '%s'" % (self.flags, self.url_string)
        _, l, _ = self.subprocess_popen(commandline)
        # Look for our files as the last element of a long list line
        return [util.fsencode(x.split()[-1]) for x in l.split(u'\n') if x and not x.startswith(u"total ")]

    def _delete(self, filename):
        commandline = u"ncftpls %s -l -X 'DELE %s' '%s'" % \
            (self.flags, filename, self.url_string)
        self.subprocess_popen(commandline)


duplicity.backend.register_backend(u"ncftp+ftp", NCFTPBackend)
duplicity.backend.uses_netloc.extend([u'ncftp+ftp'])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright (c) 2015 Matthew Bentley
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from future import standard_library
standard_library.install_aliases()
from builtins import object

from urllib.parse import quote_plus  # pylint: disable=import-error

from duplicity import log
from duplicity import progress
from duplicity import util
from duplicity import config
from duplicity.errors import BackendException, FatalBackendException
import duplicity.backend


class B2ProgressListener(object):
    def __enter__(self):
        pass

    def set_total_bytes(self, total_byte_count):
        self.total_byte_count = total_byte_count

    def bytes_completed(self, byte_count):
        progress.report_transfer(byte_count, self.total_byte_count)

    def close(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class B2Backend(duplicity.backend.Backend):
    u"""
    Backend for BackBlaze's B2 storage service
    """

    def __init__(self, parsed_url):
        u"""
        Authorize to B2 api and set up needed variables
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        global DownloadDestLocalFile, FileVersionInfoFactory

        try:  # figure out what version of b2sdk we have
            from b2sdk import __version__ as VERSION
            v_split = VERSION.split(u'.')
            self.v_num = [int(x) for x in v_split]
        except:
            self.v_num = [0, 0, 0]

        try:  # public API v2 is recommended, if available
            from b2sdk.v2 import B2Api
            from b2sdk.v2 import InMemoryAccountInfo
            from b2sdk.v2.exception import NonExistentBucket
        except ImportError:
            try:  # if public API v2 not found, try to use public API v1
                from b2sdk.v1 import B2Api
                from b2sdk.v1 import InMemoryAccountInfo
                from b2sdk.v1 import DownloadDestLocalFile
                from b2sdk.v1.exception import NonExistentBucket

                if self.v_num < [1, 9, 0]:
                    from b2sdk.v1.file_version import FileVersionInfoFactory
            except ImportError:
                try:  # try to import the new b2sdk internal API if available (and public API isn't)
                    from b2sdk.api import B2Api
                    from b2sdk.account_info import InMemoryAccountInfo
                    from b2sdk.download_dest import DownloadDestLocalFile
                    from b2sdk.exception import NonExistentBucket
                    from b2sdk.file_version import FileVersionInfoFactory
                except ImportError as e:
                    if u'b2sdk' in getattr(e, u'name', u'b2sdk'):
                        raise
                    try:  # fall back to import the old b2 client
                        from b2.api import B2Api
                        from b2.account_info import InMemoryAccountInfo
                        from b2.download_dest import DownloadDestLocalFile
                        from b2.exception import NonExistentBucket
                        from b2.file_version import FileVersionInfoFactory
                    except ImportError:
                        if u'b2' in getattr(e, u'name', u'b2'):
                            raise
                        raise BackendException(u'B2 backend requires B2 Python SDK (pip install b2sdk)')

        self.service = B2Api(InMemoryAccountInfo())
        self.parsed_url.hostname = u'B2'

        account_id = parsed_url.username
        account_key = self.get_password()

        self.url_parts = [
            x for x in parsed_url.path.replace(u"@", u"/").split(u'/') if x != u''
        ]
        if self.url_parts:
            self.username = self.url_parts.pop(0)
            bucket_name = self.url_parts.pop(0)
        else:
            raise BackendException(u"B2 requires a bucket name")
        self.path = u"".join([url_part + u"/" for url_part in self.url_parts])
        self.service.authorize_account(u'production', account_id, account_key)

        try:
            log.Log(u"B2 Backend (path= %s, bucket= %s, recommended_part_size= %s)" %
                    (self.path, bucket_name, self.service.account_info.get_recommended_part_size()), log.INFO)
        except AttributeError:
            log.Log(u"B2 Backend (path= %s, bucket= %s, minimum_part_size= %s)" %
                    (self.path, bucket_name, self.service.account_info.get_minimum_part_size()), log.INFO)

        try:
            self.bucket = self.service.get_bucket_by_name(bucket_name)
            log.Log(u"Bucket found", log.INFO)
        except NonExistentBucket:
            try:
                log.Log(u"Bucket not found, creating one", log.INFO)
                self.bucket = self.service.create_bucket(bucket_name, u'allPrivate')
            except:
                raise FatalBackendException(u"Bucket cannot be created")

    def _get(self, remote_filename, local_path):
        u"""
        Download remote_filename to local_path
        """
        log.Log(u"Get: %s -> %s" % (self.path + util.fsdecode(remote_filename),
                                    util.fsdecode(local_path.name)),
                log.INFO)
        if self.v_num < [1, 11, 0]:
            self.bucket.download_file_by_name(quote_plus(self.path + util.fsdecode(remote_filename), u'/'),
                                              DownloadDestLocalFile(local_path.name))
        else:
            df = self.bucket.download_file_by_name(quote_plus(self.path + util.fsdecode(remote_filename), u'/'))
            df.save_to(local_path.name)

    def _put(self, source_path, remote_filename):
        u"""
        Copy source_path to remote_filename
        """
        log.Log(u"Put: %s -> %s" % (util.fsdecode(source_path.name),
                                    self.path + util.fsdecode(remote_filename)),
                log.INFO)
        self.bucket.upload_local_file(util.fsdecode(source_path.name),
                                      quote_plus(self.path + util.fsdecode(remote_filename), u'/'),
                                      content_type=u'application/pgp-encrypted',
                                      progress_listener=B2ProgressListener())

    def _list(self):
        u"""
        List files on remote server
        """
        return [file_version_info.file_name[len(self.path):]
                for (file_version_info, folder_name) in self.bucket.ls(self.path)]

    def _delete(self, filename):
        u"""
        Delete filename from remote server
        """
        full_filename = self.path + util.fsdecode(filename)
        log.Log(u"Delete: %s" % full_filename, log.INFO)

        if config.b2_hide_files:
            self.bucket.hide_file(full_filename)
        else:
            file_version_info = self.file_info(quote_plus(full_filename, u'/'))
            self.bucket.delete_file_version(file_version_info.id_, file_version_info.file_name)

    def _query(self, filename):
        u"""
        Get size info of filename
        """
        log.Log(u"Query: %s" % self.path + util.fsdecode(filename), log.INFO)
        file_version_info = self.file_info(quote_plus(self.path + util.fsdecode(filename), u'/'))
        return {u'size': int(file_version_info.size)
                if file_version_info is not None and file_version_info.size is not None else -1}

    def file_info(self, filename):
        if self.v_num >= [1, 9, 0]:
            return self.bucket.get_file_info_by_name(filename)
        else:
            response = self.bucket.api.session.list_file_names(self.bucket.id_, filename, 1, self.path)
            for entry in response[u'files']:
                file_version_info = FileVersionInfoFactory.from_api_response(entry)
                if file_version_info.file_name == filename:
                    return file_version_info
            raise BackendException(u'File not found')


duplicity.backend.register_backend(u"b2", B2Backend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import os
import sys
sys.path.insert(0, os.path.abspath('../bin'))
sys.path.insert(0, os.path.abspath('../duplicity'))
sys.path.insert(0, os.path.abspath('../testing'))
sys.path.insert(0, os.path.abspath('../tools'))


# -- Project information -----------------------------------------------------

project = 'duplicity'
copyright = '2021, Kenneth Loafman'  # pylint: disable=redefined-builtin
author = 'Kenneth Loafman'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'myst_parser',
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    'testing/manual',
    'testing/override',
    'setup.py',
]


# -- Options for HTML output -------------------------------------------------

# The document name of the “master” document, that is, the document that
# contains the root toctree directive. Default is 'index' now, however,
# 'content' is still used at readthedocs.
master_doc = 'index'

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []


# -- Extension configuration -------------------------------------------------
autodoc_mock_imports = [
    'duplicity._librsync',
    'duplicity.apsw',
]

autodoc_default_options = {
    'autofunction': True,
    'members': True,
    'member-order': 'alphabetical',
    'special-members': '__init__',
    'undoc-members': True,
}

source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}


# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import os
import sys
import pytest

@pytest.fixture(scope=u"function")
def redirect_stdin():
    u"""GPG requires stdin to be open and have real file descriptor, which interferes with pytest's capture facility.
    Work around this by redirecting /dev/null to stdin temporarily.

    Activate this fixture on unittest test methods and classes by means of @pytest.mark.usefixtures("redirect_stdin")."""
    try:
        targetfd_save = os.dup(0)
        stdin_save = sys.stdin

        nullfile = open(os.devnull, u"r")
        sys.stdin = nullfile
        os.dup2(nullfile.fileno(), 0)
        yield
    finally:
        os.dup2(targetfd_save, 0)
        sys.stdin = stdin_save
        os.close(targetfd_save)
        nullfile.close()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2009 Michael Terry <mike@mterry.name>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os
import subprocess
import atexit
import signal

import duplicity.backend
from duplicity import log
from duplicity import util


def ensure_dbus():
    # GIO requires a dbus session bus which can start the gvfs daemons
    # when required.  So we make sure that such a bus exists and that our
    # environment points to it.
    if u'DBUS_SESSION_BUS_ADDRESS' not in os.environ:
        p = subprocess.Popen([u'dbus-launch'], stdout=subprocess.PIPE, universal_newlines=True)
        output = p.communicate()[0]
        lines = output.split(u'\n')
        for line in lines:
            parts = line.split(u'=', 1)
            if len(parts) == 2:
                if parts[0] == u'DBUS_SESSION_BUS_PID':  # cleanup at end
                    atexit.register(os.kill, int(parts[1]), signal.SIGTERM)
                os.environ[parts[0]] = parts[1]


class GIOBackend(duplicity.backend.Backend):
    u"""Use this backend when saving to a GIO URL.
       This is a bit of a meta-backend, in that it can handle multiple schemas.
       URLs look like schema://user@server/path.
    """
    def __init__(self, parsed_url):
        from gi.repository import Gio   # pylint: disable=import-error
        from gi.repository import GLib   # pylint: disable=import-error

        class DupMountOperation(Gio.MountOperation):
            u"""A simple MountOperation that grabs the password from the environment
               or the user.
            """
            def __init__(self, backend):
                Gio.MountOperation.__init__(self)
                self.backend = backend
                self.connect(u'ask-password', self.ask_password_cb)
                self.connect(u'ask-question', self.ask_question_cb)

            def ask_password_cb(self, *args, **kwargs):
                self.set_password(self.backend.get_password())
                self.reply(Gio.MountOperationResult.HANDLED)

            def ask_question_cb(self, *args, **kwargs):
                # Obviously just always answering with the first choice is a naive
                # approach.  But there's no easy way to allow for answering questions
                # in duplicity's typical run-from-cron mode with environment variables.
                # And only a couple gvfs backends ask questions: 'sftp' does about
                # new hosts and 'afc' does if the device is locked.  0 should be a
                # safe choice.
                self.set_choice(0)
                self.reply(Gio.MountOperationResult.HANDLED)

        duplicity.backend.Backend.__init__(self, parsed_url)

        ensure_dbus()

        self.remote_file = Gio.File.new_for_uri(parsed_url.url_string)

        # Now we make sure the location is mounted
        op = DupMountOperation(self)
        loop = GLib.MainLoop()
        self.remote_file.mount_enclosing_volume(Gio.MountMountFlags.NONE,
                                                op, None,
                                                self.__done_with_mount, loop)
        loop.run()  # halt program until we're done mounting

        # Now make the directory if it doesn't exist
        try:
            self.remote_file.make_directory_with_parents(None)
        except GLib.GError as e:
            if e.code != Gio.IOErrorEnum.EXISTS:
                raise

    def __done_with_mount(self, fileobj, result, loop):
        from gi.repository import Gio   # pylint: disable=import-error
        from gi.repository import GLib   # pylint: disable=import-error
        try:
            fileobj.mount_enclosing_volume_finish(result)
        except GLib.GError as e:
            # check for NOT_SUPPORTED because some schemas (e.g. file://) validly don't
            if e.code != Gio.IOErrorEnum.ALREADY_MOUNTED and e.code != Gio.IOErrorEnum.NOT_SUPPORTED:
                log.FatalError(_(u"Connection failed, please check your password: %s")
                               % util.uexc(e), log.ErrorCode.connection_failed)
        loop.quit()

    def __copy_progress(self, *args, **kwargs):
        pass

    def __copy_file(self, source, target):
        from gi.repository import Gio   # pylint: disable=import-error
        # Don't pass NOFOLLOW_SYMLINKS here. Some backends (e.g. google-drive:)
        # use symlinks internally for all files. In the normal course of
        # events, we never deal with symlinks anyway, just tarballs.
        source.copy(target,
                    Gio.FileCopyFlags.OVERWRITE,
                    None, self.__copy_progress, None)

    def _error_code(self, operation, e):
        from gi.repository import Gio   # pylint: disable=import-error
        from gi.repository import GLib   # pylint: disable=import-error
        if isinstance(e, GLib.GError):
            if e.code == Gio.IOErrorEnum.FAILED and operation == u'delete':
                # Sometimes delete will return a generic failure on a file not
                # found (notably the FTP does that)
                return log.ErrorCode.backend_not_found
            elif e.code == Gio.IOErrorEnum.PERMISSION_DENIED:
                return log.ErrorCode.backend_permission_denied
            elif e.code == Gio.IOErrorEnum.NOT_FOUND:
                return log.ErrorCode.backend_not_found
            elif e.code == Gio.IOErrorEnum.NO_SPACE:
                return log.ErrorCode.backend_no_space

    def _put(self, source_path, remote_filename):
        from gi.repository import Gio   # pylint: disable=import-error
        source_file = Gio.File.new_for_path(source_path.name)
        target_file = self.remote_file.get_child_for_display_name(util.fsdecode(remote_filename))
        self.__copy_file(source_file, target_file)

    def _get(self, filename, local_path):
        from gi.repository import Gio   # pylint: disable=import-error
        source_file = self.remote_file.get_child_for_display_name(util.fsdecode(filename))
        target_file = Gio.File.new_for_path(local_path.name)
        self.__copy_file(source_file, target_file)

    def _list(self):
        from gi.repository import Gio   # pylint: disable=import-error
        files = []
        # We grab display name, rather than file name because some backends
        # (e.g. google-drive:) use filesystem-specific IDs as file names and
        # only expose the "normal" name as display names. We need the display
        # name, because we try to parse them.
        enum = self.remote_file.enumerate_children(Gio.FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
                                                   Gio.FileQueryInfoFlags.NONE,
                                                   None)
        info = enum.next_file(None)
        while info:
            files.append(info.get_display_name())
            info = enum.next_file(None)
        return files

    def _delete(self, filename):
        target_file = self.remote_file.get_child_for_display_name(util.fsdecode(filename))
        target_file.delete(None)

    def _query(self, filename):
        from gi.repository import Gio   # pylint: disable=import-error
        target_file = self.remote_file.get_child_for_display_name(util.fsdecode(filename))
        info = target_file.query_info(Gio.FILE_ATTRIBUTE_STANDARD_SIZE,
                                      Gio.FileQueryInfoFlags.NONE, None)
        return {u'size': info.get_size()}


duplicity.backend.register_backend_prefix(u'gio', GIOBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""Produce and parse the names of duplicity's backup files"""

from builtins import str
from builtins import range
from builtins import object
import re
from duplicity import dup_time
from duplicity import config
import sys

full_vol_re = None
full_vol_re_short = None
full_manifest_re = None
full_manifest_re_short = None
inc_vol_re = None
inc_vol_re_short = None
inc_manifest_re = None
inc_manifest_re_short = None
full_sig_re = None
full_sig_re_short = None
new_sig_re = None
new_sig_re_short = None


def prepare_regex(force=False):
    global full_vol_re
    global full_vol_re_short
    global full_manifest_re
    global full_manifest_re_short
    global inc_vol_re
    global inc_vol_re_short
    global inc_manifest_re
    global inc_manifest_re_short
    global full_sig_re
    global full_sig_re_short
    global new_sig_re
    global new_sig_re_short

    # we force regex re-generation in unit tests because file prefixes might have changed
    if full_vol_re and not force:
        return

    full_vol_re = re.compile(b"^" + config.file_prefix + config.file_prefix_archive + b"duplicity-full"
                             b"\\.(?P<time>.*?)"
                             b"\\.vol(?P<num>[0-9]+)"
                             b"\\.difftar"
                             b"(?P<partial>(\\.part))?"
                             b"($|\\.)")

    full_vol_re_short = re.compile(b"^" + config.file_prefix + config.file_prefix_archive + b"df"
                                   b"\\.(?P<time>[0-9a-z]+?)"
                                   b"\\.(?P<num>[0-9a-z]+)"
                                   b"\\.dt"
                                   b"(?P<partial>(\\.p))?"
                                   b"($|\\.)")

    full_manifest_re = re.compile(b"^" + config.file_prefix + config.file_prefix_manifest + b"duplicity-full"
                                  b"\\.(?P<time>.*?)"
                                  b"\\.manifest"
                                  b"(?P<partial>(\\.part))?"
                                  b"($|\\.)")

    full_manifest_re_short = re.compile(b"^" + config.file_prefix + config.file_prefix_manifest + b"df"
                                        b"\\.(?P<time>[0-9a-z]+?)"
                                        b"\\.m"
                                        b"(?P<partial>(\\.p))?"
                                        b"($|\\.)")

    inc_vol_re = re.compile(b"^" + config.file_prefix + config.file_prefix_archive + b"duplicity-inc"
                            b"\\.(?P<start_time>.*?)"
                            b"\\.to\\.(?P<end_time>.*?)"
                            b"\\.vol(?P<num>[0-9]+)"
                            b"\\.difftar"
                            b"($|\\.)")

    inc_vol_re_short = re.compile(b"^" + config.file_prefix + config.file_prefix_archive + b"di"
                                  b"\\.(?P<start_time>[0-9a-z]+?)"
                                  b"\\.(?P<end_time>[0-9a-z]+?)"
                                  b"\\.(?P<num>[0-9a-z]+)"
                                  b"\\.dt"
                                  b"($|\\.)")

    inc_manifest_re = re.compile(b"^" + config.file_prefix + config.file_prefix_manifest + b"duplicity-inc"
                                 b"\\.(?P<start_time>.*?)"
                                 b"\\.to"
                                 b"\\.(?P<end_time>.*?)"
                                 b"\\.manifest"
                                 b"(?P<partial>(\\.part))?"
                                 b"(\\.|$)")

    inc_manifest_re_short = re.compile(b"^" + config.file_prefix + config.file_prefix_manifest + b"di"
                                       b"\\.(?P<start_time>[0-9a-z]+?)"
                                       b"\\.(?P<end_time>[0-9a-z]+?)"
                                       b"\\.m"
                                       b"(?P<partial>(\\.p))?"
                                       b"(\\.|$)")

    full_sig_re = re.compile(b"^" + config.file_prefix + config.file_prefix_signature + b"duplicity-full-signatures"
                             b"\\.(?P<time>.*?)"
                             b"\\.sigtar"
                             b"(?P<partial>(\\.part))?"
                             b"(\\.|$)")

    full_sig_re_short = re.compile(b"^" + config.file_prefix + config.file_prefix_signature + b"dfs"
                                   b"\\.(?P<time>[0-9a-z]+?)"
                                   b"\\.st"
                                   b"(?P<partial>(\\.p))?"
                                   b"(\\.|$)")

    new_sig_re = re.compile(b"^" + config.file_prefix + config.file_prefix_signature + b"duplicity-new-signatures"
                            b"\\.(?P<start_time>.*?)"
                            b"\\.to"
                            b"\\.(?P<end_time>.*?)"
                            b"\\.sigtar"
                            b"(?P<partial>(\\.part))?"
                            b"(\\.|$)")

    new_sig_re_short = re.compile(b"^" + config.file_prefix + config.file_prefix_signature + b"dns"
                                  b"\\.(?P<start_time>[0-9a-z]+?)"
                                  b"\\.(?P<end_time>[0-9a-z]+?)"
                                  b"\\.st"
                                  b"(?P<partial>(\\.p))?"
                                  b"(\\.|$)")


def to_base36(n):
    u"""
    Return string representation of n in base 36 (use 0-9 and a-z)
    """
    div, mod = divmod(n, 36)
    if mod <= 9:
        last_digit = str(mod)
    else:
        last_digit = chr(ord(u'a') + mod - 10)
    if sys.version_info.major >= 3:
        last_digit = last_digit.encode()
    if n == mod:
        return last_digit
    else:
        return to_base36(div) + last_digit


def from_base36(s):
    u"""
    Convert string s in base 36 to long int
    """
    total = 0
    for i in range(len(s)):
        total *= 36
        if sys.version_info.major >= 3 and isinstance(s, bytes):
            digit_ord = s[i]
        else:
            digit_ord = ord(s[i])
        if ord(u'0') <= digit_ord <= ord(u'9'):
            total += digit_ord - ord(u'0')
        elif ord(u'a') <= digit_ord <= ord(u'z'):
            total += digit_ord - ord(u'a') + 10
        else:
            assert 0, u"Digit %s in %s not in proper range" % (s[i], s)
    return total


def get_suffix(encrypted, gzipped):
    u"""
    Return appropriate suffix depending on status of
    encryption, compression, and short_filenames.
    """
    if encrypted:
        gzipped = False
    if encrypted:
        if config.short_filenames:
            suffix = b'.g'
        else:
            suffix = b".gpg"
    elif gzipped:
        if config.short_filenames:
            suffix = b".z"
        else:
            suffix = b'.gz'
    else:
        suffix = b""
    return suffix


def get(type, volume_number=None, manifest=False,  # pylint: disable=redefined-builtin
        encrypted=False, gzipped=False, partial=False):
    u"""
    Return duplicity filename of specified type

    type can be "full", "inc", "full-sig", or "new-sig". volume_number
    can be given with the full and inc types.  If manifest is true the
    filename is of a full or inc manifest file.
    """
    assert dup_time.curtimestr
    if encrypted:
        gzipped = False
    suffix = get_suffix(encrypted, gzipped)
    part_string = b""
    if config.short_filenames:
        if partial:
            part_string = b".p"
    else:
        if partial:
            part_string = b".part"

    if type == u"full-sig" or type == u"new-sig":
        assert not volume_number and not manifest
        assert not (volume_number and part_string)
        if type == u"full-sig":
            if config.short_filenames:
                return (config.file_prefix + config.file_prefix_signature +
                        b"dfs.%s.st%s%s" %
                        (to_base36(dup_time.curtime), part_string, suffix))
            else:
                return (config.file_prefix + config.file_prefix_signature +
                        b"duplicity-full-signatures.%s.sigtar%s%s" %
                        (dup_time.curtimestr.encode(), part_string, suffix))
        elif type == u"new-sig":
            if config.short_filenames:
                return (config.file_prefix + config.file_prefix_signature +
                        b"dns.%s.%s.st%s%s" %
                        (to_base36(dup_time.prevtime),
                         to_base36(dup_time.curtime),
                         part_string, suffix))
            else:
                return (config.file_prefix + config.file_prefix_signature +
                        b"duplicity-new-signatures.%s.to.%s.sigtar%s%s" %
                        (dup_time.prevtimestr.encode(), dup_time.curtimestr.encode(),
                         part_string, suffix))
    else:
        assert volume_number or manifest
        assert not (volume_number and manifest)

        prefix = config.file_prefix

        if volume_number:
            if config.short_filenames:
                vol_string = b"%s.dt" % to_base36(volume_number)
            else:
                vol_string = b"vol%d.difftar" % volume_number
            prefix += config.file_prefix_archive
        else:
            if config.short_filenames:
                vol_string = b"m"
            else:
                vol_string = b"manifest"
            prefix += config.file_prefix_manifest

        if type == u"full":
            if config.short_filenames:
                return (b"%sdf.%s.%s%s%s" % (prefix, to_base36(dup_time.curtime),
                                             vol_string, part_string, suffix))
            else:
                return (b"%sduplicity-full.%s.%s%s%s" % (prefix, dup_time.curtimestr.encode(),
                                                         vol_string, part_string, suffix))
        elif type == u"inc":
            if config.short_filenames:
                return (b"%sdi.%s.%s.%s%s%s" % (prefix, to_base36(dup_time.prevtime),
                                                to_base36(dup_time.curtime),
                                                vol_string, part_string, suffix))
            else:
                return (b"%sduplicity-inc.%s.to.%s.%s%s%s" % (prefix, dup_time.prevtimestr.encode(),
                                                              dup_time.curtimestr.encode(),
                                                              vol_string, part_string, suffix))
        else:
            assert 0


def parse(filename):
    u"""
    Parse duplicity filename, return None or ParseResults object
    """
    def str2time(timestr, short):
        u"""
        Return time in seconds if string can be converted, None otherwise
        """
        if isinstance(timestr, bytes):
            timestr = timestr.decode()

        if short:
            t = from_base36(timestr)
        else:
            try:
                t = dup_time.genstrtotime(timestr.upper())
            except dup_time.TimeException:
                return None
        return t

    def get_vol_num(s, short):
        u"""
        Return volume number from volume number string
        """
        if short:
            return from_base36(s)
        else:
            return int(s)

    def check_full():
        u"""
        Return ParseResults if file is from full backup, None otherwise
        """
        prepare_regex()
        short = True
        m1 = full_vol_re_short.search(filename)
        m2 = full_manifest_re_short.search(filename)
        if not m1 and not m2 and not config.short_filenames:
            short = False
            m1 = full_vol_re.search(filename)
            m2 = full_manifest_re.search(filename)
        if m1 or m2:
            t = str2time((m1 or m2).group(u"time"), short)
            if t:
                if m1:
                    return ParseResults(u"full", time=t,
                                        volume_number=get_vol_num(m1.group(u"num"), short))
                else:
                    return ParseResults(u"full", time=t, manifest=True,
                                        partial=(m2.group(u"partial") is not None))
        return None

    def check_inc():
        u"""
        Return ParseResults if file is from inc backup, None otherwise
        """
        prepare_regex()
        short = True
        m1 = inc_vol_re_short.search(filename)
        m2 = inc_manifest_re_short.search(filename)
        if not m1 and not m2 and not config.short_filenames:
            short = False
            m1 = inc_vol_re.search(filename)
            m2 = inc_manifest_re.search(filename)
        if m1 or m2:
            t1 = str2time((m1 or m2).group(u"start_time"), short)
            t2 = str2time((m1 or m2).group(u"end_time"), short)
            if t1 and t2:
                if m1:
                    return ParseResults(u"inc", start_time=t1,
                                        end_time=t2, volume_number=get_vol_num(m1.group(u"num"), short))
                else:
                    return ParseResults(u"inc", start_time=t1, end_time=t2, manifest=1,
                                        partial=(m2.group(u"partial") is not None))
        return None

    def check_sig():
        u"""
        Return ParseResults if file is a signature, None otherwise
        """
        prepare_regex()
        short = True
        m = full_sig_re_short.search(filename)
        if not m and not config.short_filenames:
            short = False
            m = full_sig_re.search(filename)
        if m:
            t = str2time(m.group(u"time"), short)
            if t:
                return ParseResults(u"full-sig", time=t,
                                    partial=(m.group(u"partial") is not None))
            else:
                return None

        short = True
        m = new_sig_re_short.search(filename)
        if not m and not config.short_filenames:
            short = False
            m = new_sig_re.search(filename)
        if m:
            t1 = str2time(m.group(u"start_time"), short)
            t2 = str2time(m.group(u"end_time"), short)
            if t1 and t2:
                return ParseResults(u"new-sig", start_time=t1, end_time=t2,
                                    partial=(m.group(u"partial") is not None))
        return None

    def set_encryption_or_compression(pr):
        u"""
        Set encryption and compression flags in ParseResults pr
        """
        if (filename.endswith(b'.z') or
                not config.short_filenames and filename.endswith(b'gz')):
            pr.compressed = 1
        else:
            pr.compressed = None

        if (filename.endswith(b'.g') or
                not config.short_filenames and filename.endswith(b'.gpg')):
            pr.encrypted = 1
        else:
            pr.encrypted = None

    pr = check_full()
    if not pr:
        pr = check_inc()
        if not pr:
            pr = check_sig()
            if not pr:
                return None
    set_encryption_or_compression(pr)
    return pr


class ParseResults(object):
    u"""
    Hold information taken from a duplicity filename
    """
    def __init__(self, type, manifest=None, volume_number=None,  # pylint: disable=redefined-builtin
                 time=None, start_time=None, end_time=None,
                 encrypted=None, compressed=None, partial=False):

        assert type in [u"full-sig", u"new-sig", u"inc", u"full"]

        self.type = type
        if type == u"inc" or type == u"full":
            assert manifest or volume_number
        if type == u"inc" or type == u"new-sig":
            assert start_time and end_time
        else:
            assert time

        self.manifest = manifest
        self.volume_number = volume_number
        self.time = time
        self.start_time, self.end_time = start_time, end_time

        self.compressed = compressed  # true if gzip compressed
        self.encrypted = encrypted  # true if gpg encrypted

        self.partial = partial

    def __eq__(self, other):
        return self.type == other.type and \
            self.manifest == other.manifest and \
            self.time == other.time and \
            self.start_time == other.start_time and \
            self.end_time == other.end_time and \
            self.partial == other.partial



#!/usr/bin/env python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2018 Aaron Whitehouse <aaron@whitehouse.kiwi.nz>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# For predictable results in python/3 all string literals need to be marked as unicode, bytes or raw
# This code finds all unadorned string literals (strings that are not marked with a u, b or r)

from __future__ import print_function

import sys
import tokenize
import token

# Unfortunately python does not have the useful named tuple result from tokenize.tokenize,
# so we have to recreate the effect using namedtuple and tokenize.generate_tokens
from collections import namedtuple
python_token = namedtuple(u'python_token', u'type string start end line')


def return_unadorned_string_tokens(f):
    if sys.version_info[0] < 3:
        unnamed_tokens = tokenize.generate_tokens(f.readline)
        for t in unnamed_tokens:
            named_token = python_token(token.tok_name[t[0]], *t[1:])
            if named_token.type == u"STRING" and named_token.string[0] in [u'"', u"'"]:
                yield named_token

    else:
        named_tokens = tokenize.tokenize(f.readline)
        for t in named_tokens:
            if t.type == token.STRING and t.string[0] in [u'"', u"'"]:
                yield t


def check_file_for_unadorned(python_file):
    unadorned_string_list = []
    with open(python_file, u'rb') as f:
        for s in return_unadorned_string_tokens(f):
            unadorned_string_list.append((python_file, s.start, s.end, s.string))
    return unadorned_string_list


if __name__ == u"__main__":
    import argparse

    parser = argparse.ArgumentParser(description=u'Fix any unadorned string literals in a Python file')
    parser.add_argument(u'file', help=u'The file to search')
    args = parser.parse_args()

    lines = open(args.file, u"r").readlines()
    lines = [list(line) for line in lines]

    unadorned_string_list = check_file_for_unadorned(args.file)
    if len(unadorned_string_list) == 0:
        print(u"There are no unadorned strings in", args.file)
    else:
        print(u"There are unadorned strings in", args.file, u"\n")

        locs = {}
        for unadorned_string in unadorned_string_list:
            print(unadorned_string)
            python_file, string_start, string_end, strings = unadorned_string
            linenum = int(string_start[0]) - 1
            colnum = int(string_start[1])
            if linenum in locs:
                locs[linenum].insert(0, colnum)
            else:
                locs[linenum] = [colnum]

        for linenum in sorted(locs.keys()):
            for colnum in locs[linenum]:
                lines[linenum].insert(colnum, u'u')
            print(u"{}: {}: {}".format(linenum, len(locs[linenum]), u''.join(lines[linenum][:-1])))

        lines = [u''.join(line) for line in lines]
        open(args.file, u"w").writelines(lines)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2011 Fabian Topfstedt <topfstedt@schneevonmorgen.com>
#
# This module is included with granted permission from the original author.
# The original source is available at http://bitbucket.org/fabian/filechunkio

import io
import os


SEEK_SET = getattr(io, u'SEEK_SET', 0)
SEEK_CUR = getattr(io, u'SEEK_CUR', 1)
SEEK_END = getattr(io, u'SEEK_END', 2)


class FileChunkIO(io.FileIO):
    u"""
    A class that allows you reading only a chunk of a file.
    """
    def __init__(self, name, mode=u'r', closefd=True, offset=0, bytes=None,  # pylint: disable=redefined-builtin
                 *args, **kwargs):  # pylint: disable=redefined-builtin
        u"""
        Open a file chunk. The mode can only be 'r' for reading. Offset
        is the amount of bytes that the chunks starts after the real file's
        first byte. Bytes defines the amount of bytes the chunk has, which you
        can set to None to include the last byte of the real file.
        """
        if not mode.startswith(u'r'):
            raise ValueError(u"Mode string must begin with 'r'")
        self.offset = offset
        self.bytes = bytes
        if bytes is None:
            self.bytes = os.stat(name).st_size - self.offset
        super(FileChunkIO, self).__init__(name, mode, closefd, *args, **kwargs)
        self.seek(0)

    def seek(self, offset, whence=SEEK_SET):
        u"""
        Move to a new chunk position.
        """
        if whence == SEEK_SET:
            super(FileChunkIO, self).seek(self.offset + offset)
        elif whence == SEEK_CUR:
            self.seek(self.tell() + offset)
        elif whence == SEEK_END:
            self.seek(self.bytes + offset)

    def tell(self):
        u"""
        Current file position.
        """
        return super(FileChunkIO, self).tell() - self.offset

    def read(self, n=-1):
        u"""
        Read and return at most n bytes.
        """
        if n >= 0:
            max_n = self.bytes - self.tell()
            n = min([n, max_n])
            return super(FileChunkIO, self).read(n)
        else:
            return self.readall()

    def readall(self):
        u"""
        Read all data from the chunk.
        """
        return self.read(self.bytes - self.tell())

    def readinto(self, b):
        u"""
        Same as RawIOBase.readinto().
        """
        data = self.read(len(b))
        n = len(data)
        try:
            b[:n] = data
        except TypeError as err:
            import array
            if not isinstance(b, array.array):
                raise err
            b[:n] = array.array(b'b', data)
        return n



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2015 Yigal Asnis
# Copyright 2021 Jindrich Makovicka
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# It is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from builtins import str

import os
import pickle

from duplicity import log
from duplicity import util
from duplicity.errors import BackendException
import duplicity.backend


class GDriveBackend(duplicity.backend.Backend):
    u"""Connect to remote store using Google Drive API V3"""

    PAGE_SIZE = 100
    MIN_RESUMABLE_UPLOAD = 5 * 1024 * 1024

    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)
        try:
            from googleapiclient.discovery import build
            from google.oauth2.service_account import Credentials
        except ImportError as e:
            raise BackendException(u"""\
GDrive backend requires Google API client installation.
Please read the manpage for setup details.
Exception: %s""" % str(e))

        # Note Google has 2 drive methods, `Shared(previously Team) Drives` and `My Drive`
        #   both can be shared but require different addressing
        # For a Google Shared Drives folder
        # ---------------------------------
        # Share Drive ID specified as a query parameter in the backend URL.
        # Example:
        #  gdrive://developer.gserviceaccount.com/target-folder/?driveID=<SHARED DRIVE ID>
        #
        # For a Google My Drive based shared folder
        # -----------------------------------------
        # MyDrive folder ID specified as a query parameter in the backend URL
        #
        # Example
        #  export GOOGLE_SERVICE_ACCOUNT_URL=<serviceaccount-name>@<serviceaccount-name>.iam.gserviceaccount.com
        #  gdrive://${GOOGLE_SERVICE_ACCOUNT_URL}/<target-folder-name/>?myDriveFolderID=<google-myDrive-folder-id>
        #
        # both methods use a Google Services Account
        # export GOOGLE_SERVICE_JSON_FILE=<serviceaccount-credentials.json>
        # export GOOGLE_SERVICE_ACCOUNT_URL=<serviceaccount-name>@<serviceaccount-name>.iam.gserviceaccount.com

        self.shared_drive_corpora = {}
        self.shared_drive_id = {}
        self.shared_drive_flags_include = {}
        self.shared_drive_flags_support = {}
        self.shared_root_folder_id = None
        if u'driveID' in parsed_url.query_args:
            self.shared_drive_corpora = {u'corpora': u'drive'}
            self.shared_drive_id = {u'driveId': parsed_url.query_args[u'driveID'][0]}
            self.shared_drive_flags_include = {u'includeItemsFromAllDrives': True}
            self.shared_drive_flags_support = {u'supportsAllDrives': True}
        elif u'myDriveFolderID' in parsed_url.query_args:
            self.shared_drive_corpora = {u'corpora': u'user'}
            self.shared_drive_flags_include = {u'includeItemsFromAllDrives': True}
            self.shared_drive_flags_support = {u'supportsAllDrives': True}
            self.shared_root_folder_id = parsed_url.query_args[u'myDriveFolderID'][0]
        else:
            raise BackendException(
                u"gdrive: backend requires a query paramater should either be driveID or myDriveFolderID")
        if parsed_url.username is not None:
            client_id = parsed_url.username + u'@' + parsed_url.hostname
        else:
            client_id = parsed_url.hostname

        if u'GOOGLE_SERVICE_JSON_FILE' in os.environ:
            credentials = Credentials.from_service_account_file(os.environ[u'GOOGLE_SERVICE_JSON_FILE'])
            if credentials.service_account_email != client_id:
                raise BackendException(
                    u'Service account email in the JSON file (%s) does not match the URL (%s)' %
                    (credentials.service_account_email, client_id))

        elif u'GOOGLE_CLIENT_SECRET_JSON_FILE' in os.environ and u'GOOGLE_CREDENTIALS_FILE' in os.environ:
            from google_auth_oauthlib.flow import InstalledAppFlow
            from google.auth.transport.requests import Request

            credentials = None
            if os.path.exists(os.environ[u'GOOGLE_CREDENTIALS_FILE']):
                with open(os.environ[u'GOOGLE_CREDENTIALS_FILE'], u'rb') as token:
                    credentials = pickle.load(token)

            # If there are no (valid) credentials available, let the user log in.
            if not credentials or not credentials.valid:
                if credentials and credentials.expired and credentials.refresh_token:
                    credentials.refresh(Request())
                else:
                    flow = InstalledAppFlow.from_client_secrets_file(
                        os.environ[u'GOOGLE_CLIENT_SECRET_JSON_FILE'],
                        [u'https://www.googleapis.com/auth/drive.file'])

                    if flow.client_config[u'client_id'] != client_id:
                        raise BackendException(
                            u'Client ID in the JSON file (%s) does not match the URL (%s)' %
                            (flow.client_config[u'client_id'], client_id))

                    credentials = flow.run_console()
                # Save the credentials for the next run
                with open(os.environ[u'GOOGLE_CREDENTIALS_FILE'], u'wb') as token:
                    pickle.dump(credentials, token)

            if credentials.client_id != client_id:
                raise BackendException(
                    u'Client ID in the credentials file (%s) does not match the URL (%s)' %
                    (credentials.client_id, client_id))

        else:
            raise BackendException(
                u'GOOGLE_SERVICE_JSON_FILE or GOOGLE_CLIENT_SECRET_JSON_FILE environment '
                u'variable not set. Please read the manpage to fix.')

        self.drive = build(u'drive', u'v3', credentials=credentials)

        if self.shared_drive_id:
            parent_folder_id = self.shared_drive_id[u'driveId']
        elif self.shared_root_folder_id:
            parent_folder_id = self.shared_root_folder_id
        else:
            parent_folder_id = u'root'

        # Fetch destination folder entry and create hierarchy if required.
        folder_names = parsed_url.path.split(u'/')
        for folder_name in folder_names:

            if not folder_name:
                continue
            q = (u"name = '" + folder_name + u"' and '" + parent_folder_id +
                 u"' in parents and mimeType = 'application/vnd.google-apps.folder' and trashed=false")
            results = self.drive.files().list(q=q,
                                              pageSize=1,
                                              fields=u"files(name,id),nextPageToken",
                                              **self.shared_drive_corpora,
                                              **self.shared_drive_id,
                                              **self.shared_drive_flags_include,
                                              **self.shared_drive_flags_support).execute()
            file_list = results.get(u'files', [])
            if len(file_list) == 0:
                file_metadata = {u'name': folder_name,
                                 u'mimeType': u"application/vnd.google-apps.folder",
                                 u'parents': [parent_folder_id]}
                file_metadata.update(self.shared_drive_id)
                folder = self.drive.files().create(body=file_metadata,
                                                   fields=u'id',
                                                   **self.shared_drive_flags_support).execute()
            else:
                folder = file_list[0]

            parent_folder_id = folder[u'id']

        self.folder = parent_folder_id
        self.id_cache = {}

    def file_by_name(self, filename):
        from googleapiclient.errors import HttpError

        filename = util.fsdecode(filename)

        if filename in self.id_cache:
            # It might since have been locally moved, renamed or deleted, so we
            # need to validate the entry.
            file_id = self.id_cache[filename]
            try:
                drive_file = self.drive.files().get(fileId=file_id,
                                                    fields=u'id,size,name,parents,trashed',
                                                    **self.shared_drive_flags_support).execute()
                if drive_file[u'name'] == filename and not drive_file[u'trashed']:
                    for parent in drive_file[u'parents']:
                        if parent == self.folder:
                            log.Info(u"GDrive backend: found file '%s' with id %s in ID cache" %
                                     (filename, file_id))
                            return drive_file
            except HttpError as error:
                # A 404 occurs if the ID is no longer valid
                if error.resp.status != 404:
                    raise
            # If we get here, the cache entry is invalid
            log.Info(u"GDrive backend: invalidating '%s' (previously ID %s) from ID cache" %
                     (filename, file_id))
            del self.id_cache[filename]

        # Not found in the cache, so use directory listing. This is less
        # reliable because there is no strong consistency.
        q = u"name = '%s' and '%s' in parents and trashed = false" % (filename, self.folder)
        results = self.drive.files().list(q=q, fields=u'files(name,id,size),nextPageToken',
                                          pageSize=2,
                                          **self.shared_drive_corpora,
                                          **self.shared_drive_id,
                                          **self.shared_drive_flags_include,
                                          **self.shared_drive_flags_support).execute()
        file_list = results.get(u'files', [])
        if len(file_list) > 1:
            log.FatalError(u"GDrive backend: multiple files called '%s'." % (filename,))
        elif len(file_list) > 0:
            file_id = file_list[0][u'id']
            self.id_cache[filename] = file_list[0][u'id']
            log.Info(u"GDrive backend: found file '%s' with id %s on server, "
                     u"adding to cache" % (filename, file_id))
            return file_list[0]

        log.Info(u"GDrive backend: file '%s' not found in cache or on server" %
                 (filename,))
        return None

    def id_by_name(self, filename):
        drive_file = self.file_by_name(filename)
        if drive_file is None:
            return u''
        else:
            return drive_file[u'id']

    def _put(self, source_path, remote_filename):
        from googleapiclient.http import MediaFileUpload

        remote_filename = util.fsdecode(remote_filename)
        drive_file = self.file_by_name(remote_filename)
        if remote_filename.endswith(u'.gpg'):
            mime_type = u'application/pgp-encrypted'
        else:
            mime_type = u'text/plain'

        file_size = os.path.getsize(source_path.name)
        if file_size >= self.MIN_RESUMABLE_UPLOAD:
            resumable = True
            num_retries = 5
        else:
            resumable = False
            num_retries = 0

        media = MediaFileUpload(source_path.name, mimetype=mime_type, resumable=resumable)
        if drive_file is None:
            # No existing file, make a new one
            file_metadata = {u'name': remote_filename, u'parents': [self.folder]}
            file_metadata.update(self.shared_drive_id)
            log.Info(u"GDrive backend: creating new file '%s'" % (remote_filename,))
            drive_file = self.drive.files().create(
                body=file_metadata,
                media_body=media,
                **self.shared_drive_flags_support).execute(num_retries=num_retries)
        else:
            log.Info(u"GDrive backend: replacing existing file '%s' with id '%s'" % (
                remote_filename, drive_file[u'id']))
            drive_file = self.drive.files().update(
                media_body=media,
                fileId=drive_file[u'id'],
                **self.shared_drive_flags_support).execute(num_retries=num_retries)

        self.id_cache[remote_filename] = drive_file[u'id']

    def _get(self, remote_filename, local_path):
        from googleapiclient.http import MediaIoBaseDownload

        drive_file = self.file_by_name(remote_filename)
        request = self.drive.files().get_media(fileId=drive_file[u'id'],
                                               **self.shared_drive_flags_support)
        with open(util.fsdecode(local_path.name), u"wb") as fh:
            done = False
            downloader = MediaIoBaseDownload(fh, request)
            while done is False:
                status, done = downloader.next_chunk()

    def _list(self):
        page_token = None
        drive_files = []
        while True:
            response = self.drive.files().list(
                q=u"'" + self.folder + u"' in parents and trashed=false",
                pageSize=self.PAGE_SIZE,
                fields=u"files(name,id),nextPageToken",
                pageToken=page_token,
                **self.shared_drive_corpora,
                **self.shared_drive_id,
                **self.shared_drive_flags_include,
                **self.shared_drive_flags_support).execute()

            drive_files += response.get(u'files', [])

            page_token = response.get(u'nextPageToken', None)
            if page_token is None:
                break

        filenames = set(item[u'name'] for item in drive_files)
        # Check the cache as well. A file might have just been uploaded but
        # not yet appear in the listing.
        # Note: do not use iterkeys() here, because file_by_name will modify
        # the cache if it finds invalid entries.
        for filename in list(self.id_cache.keys()):
            if (filename not in filenames) and (self.file_by_name(filename) is not None):
                filenames.add(filename)
        return list(filenames)

    def _delete(self, filename):
        file_id = self.id_by_name(filename)
        if file_id == u'':
            log.Warn(u"File '%s' does not exist while trying to delete it" % (util.fsdecode(filename),))
        else:
            self.drive.files().delete(fileId=file_id,
                                      **self.shared_drive_flags_support).execute()

    def _query(self, filename):
        drive_file = self.file_by_name(filename)
        if drive_file is None:
            size = -1
        else:
            size = int(drive_file[u'size'])
        return {u'size': size}

    def _error_code(self, operation, error):  # pylint: disable=unused-argument
        from google.auth.exceptions import RefreshError
        from googleapiclient.errors import HttpError
        if isinstance(error, HttpError):
            return log.ErrorCode.backend_not_found
        elif isinstance(error, RefreshError):
            return log.ErrorCode.backend_permission_denied
        return log.ErrorCode.backend_error


duplicity.backend.register_backend(u'gdrive', GDriveBackend)

duplicity.backend.uses_netloc.extend([u'gdrive'])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Menno Smits <menno@smi-ling.nl>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os
import urllib
import tempfile
import re
import xml.etree.ElementTree as ET
import shutil
import errno


import duplicity.backend
from duplicity import config
from duplicity import log
from duplicity import tempdir
from duplicity import progress
from duplicity.errors import BackendException

#
#   This backend works with the IDrive  "dedup implementation". V0.1
#               (for all new and recent accounts)
#
#   Credits: This code is loosely inspired by the work of <aappddeevv>
#
#
#   This backend uses an intermediate driver for IDrive: "idevsutil_dedup" that will be
#   installed automagically  when you perform the account setup on your system.
#   It can, however, also be downloaded directly from the following URL's
#
#   https://www.idrivedownloads.com/downloads/linux/download-options/IDrive_linux_64bit.zip
#   and
#   https://www.idrivedownloads.com/downloads/linux/download-options/IDrive_linux_32bit.zip
#
#   for 32 and 64 bit linux, respectively. Copy the file anywhere with exe permissions.
#   (no further setup of your IDrive account is needed for idrived to work)
#
#
#   For this backend to work, you need to create a number of environment variables:
#
#   - Put the absolute path to the driver-file (idevsutil_dedup) in IDEVSPATH
#   - Put the account-name (login name) in IDRIVEID
#
#   - Put the name of the desired bucket for this backup-session in IDBUCKET
#     If this bucket does not exist it will be created at runtime
#
#   - Create a file with the account password - put absolute path in IDPWDFILE
#
#   When using a custom encryption key:
#   - Create a file with the encryption key - put absolute path in IDKEYFILE
#
#   Note: setup proper security for these files!
#
#
#   The IDrive "root" issue ...
#
#   IDrive stores files according to 1) the selected bucket, 2) the supplied path
#   and 3)the absolute path of the directory used for uploads. So ... if we use
#       - bucket <MYBUCKET>
#       - duplicity commandline idrived://DUPLICITY
#   and
#       - system tempfile OR path set from --tempfile "\tmp"
#
#   the files end-up in the following path:
#       <MYBUCKET>/DUPLICITY/tmp/duplicity-??????-tempdir
#
#   Not only is this SO UGLY .... but - as tempdirs have unique names - this effectively
#   disables the idea of incremental backups.
#
#   To remedy this, idrived uses the concept of a "fakeroot" directory, defined via the
#   --idr-fakeroot=... switch. This can be an existing directory, or the directory is
#   created at runtime on the root of the (host) files system. (cave: you have to have
#   write access to the root!). Directories created at runtime are auto-removed on exit!
#
#   So, in the above scheme, we could do:
#       duplicity --idr-fakeroot=nicepath idrived://DUPLICITY
#
#   our files end-up at
#       <MYBUCKET>/DUPLICITY/nicepath
#
#
#   Have fun!
#


class IDriveBackend(duplicity.backend.Backend):

    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        # parsed_url will have leading slashes in it, 4 slashes typically.
        self.parsed_url = parsed_url
        self.url_string = duplicity.backend.strip_auth_from_url(self.parsed_url)
        log.Debug(u"parsed_url: {0}".format(parsed_url))

        self.connected = False

    def user_connected(self):
        return self.connected

    def request(self, commandline):
        # request for commands returning data in XML format
        log.Debug(u"Request command: {0}".format(commandline))
        try:
            _, reply, error = self.subprocess_popen(commandline)
        except KeyError:
            raise BackendException(u"Unknown protocol failure on request {0}".format(commandline))

        response = reply + error
        try:
            xml = u"<root>" + u''.join(re.findall(u"<[^>]+>", response)) + u"</root>"
            el = ET.fromstring(xml)

        except:
            el = None
        log.Debug(u"Request response: {0}".format(response))

        return el

    def connect(self):
        # get the path to the command executable
        path = os.environ.get(u"IDEVSPATH")
        if path is None:
            log.Warn(u"-" * 72)
            log.Warn(u"WARNING: No path to 'idevsutil_dedup' has been set. Download module from")
            log.Warn(u"   https://www.idrivedownloads.com/downloads/linux/download-options/IDrive_linux_64bit.zip")
            log.Warn(u"or")
            log.Warn(u"   https://www.idrivedownloads.com/downloads/linux/download-options/IDrive_linux_32bit.zip")
            log.Warn(u"and place anywhere with exe rights. Then creat env var 'IDEVSPATH' with path to file")
            log.Warn(u"-" * 72)
            raise BackendException(u"No IDEVSPATH env var set. Should contain folder to idevsutil_dedup")
        self.cmd = os.path.join(path, u"idevsutil_dedup")
        log.Debug(u"IDrive command base: %s" % (self.cmd))

        # get the account-id
        self.idriveid = os.environ.get(u"IDRIVEID")
        if self.idriveid is None:
            log.Warn(u"-" * 72)
            log.Warn(u"WARNING: IDrive logon ID missing")
            log.Warn(u"Create an environment variable IDriveID with your IDrive logon ID")
            log.Warn(u"-" * 72)
            raise BackendException(u"No IDRIVEID env var set. Should contain IDrive id")
        log.Debug(u"IDrive id: %s" % (self.idriveid))

        # Get the full-path to the account password file
        filepath = os.environ.get(u"IDPWDFILE")
        if filepath is None:
            log.Warn(u"-" * 72)
            log.Warn(u"WARNING: IDrive password file missging")
            log.Warn(u"Please create a file with your IDrive logon password,")
            log.Warn(u"Then create an environment variable IDPWDFILE with path/filename of said file")
            log.Warn(u"-" * 72)
            raise BackendException(u"No IDPWDFILE env var set. Should contain file with password")
        log.Debug(u"IDrive pwdpath: %s" % (filepath))
        self.auth_switch = u" --password-file={0}".format(filepath)

        # fakeroot set? Create directory and mark for cleanup
        if config.fakeroot is None:
            self.cleanup = False
            self.fakeroot = u''
        else:
            # Make sure fake root is created at root level!
            self.fakeroot = os.path.join(u'/', config.fakeroot)
            try:
                os.mkdir(self.fakeroot)
            except OSError as e:
                self.cleanup = False
                if e.errno == errno.EEXIST:
                    log.Debug(u"Using existing directory {0} as fake-root".format(self.fakeroot))
                else:
                    log.Warn(u"-" * 72)
                    log.Warn(u"WARNING: Creation of FAKEROOT {0} failed; backup will use system temp directory"
                             .format(self.fakeroot))
                    log.Warn(u"This might interfere with incremental backups")
                    log.Warn(u"-" * 72)
                    raise BackendException(u"Creation of the directory {0} failed".format(self.fakeroot))
            else:
                log.Debug(u"Directory {0} created as fake-root (Will clean-up afterwards!)".format(self.fakeroot))
                self.cleanup = True

        # get the bucket
        self.bucket = os.environ.get(u"IDBUCKET")
        if self.bucket is None:
            log.Warn(u"-" * 72)
            log.Warn(u"WARNING: IDrive backup bucket missing")
            log.Warn(u"Create an environment variable IDBUCKET specifying the target bucket")
            log.Warn(u"-" * 72)
            raise BackendException(u"No IDBUCKET env var set. Should contain IDrive backup bucket")
        log.Debug(u"IDrive bucket: %s" % (self.bucket))

        # check account / get config status and config type
        el = self.request(self.cmd + self.auth_switch + u" --validate --user={0}".format(self.idriveid)).find(u'tree')

        if el.attrib[u"message"] != u"SUCCESS":
            raise BackendException(u"Protocol failure - " + el.attrib[u"desc"])
        if el.attrib[u"desc"] != u"VALID ACCOUNT":
            raise BackendException(u"IDrive account invalid")
        if el.attrib[u"configstatus"] != u"SET":
            raise BackendException(u"IDrive account not set")

        # When private encryption enabled: get the full-path to a encription key file
        if el.attrib[u"configtype"] == u"PRIVATE":
            filepath = os.environ.get(u"IDKEYFILE")
            if filepath is None:
                log.Warn(u"-" * 72)
                log.Warn(u"WARNING: IDrive encryption key file missging")
                log.Warn(u"Please create a file with your IDrive encryption key,")
                log.Warn(u"Then create an environment variable IDKEYFILE with path/filename of said file")
                log.Warn(u"-" * 72)
                raise BackendException(u"No IDKEYFILE env var set. Should contain file with encription key")
            log.Debug(u"IDrive keypath: %s" % (filepath))
            self.auth_switch += u" --pvt-key={0}".format(filepath)

        # get the server address
        el = self.request(self.cmd + self.auth_switch + u" --getServerAddress {0}".format(self.idriveid)).find(u'tree')
        self.idriveserver = el.attrib[u"cmdUtilityServer"]

        # get the device list - primarely used to get device-id string
        el = self.request(self.cmd + self.auth_switch + u" --list-device {0}@{1}::home".
                          format(self.idriveid, self.idriveserver))
        # scan all returned devices for requested device (== bucket)
        self.idrivedevid = None
        for item in el.findall(u'item'):
            if item.attrib[u'nick_name'] == self.bucket:
                # prefix and suffix reverse-engineered from Common.pl!
                self.idrivedevid = u"5c0b" + item.attrib[u"device_id"] + u"4b5z"
        if self.idrivedevid is None:
            el = self.request(
                self.cmd + self.auth_switch +
                u" --create-bucket --bucket-type=D --nick-name={0} --os=Linux --uid=987654321 {1}@{2}::home/"
                .format(self.bucket, self.idriveid, self.idriveserver)).find(u'item')
            # prefix and suffix reverse-engineered from Common.pl!
            self.idrivedevid = u"5c0b" + el.attrib[u"device_id"] + u"4b5z"

        # We're fully connected!
        self.connected = True
        log.Debug(u"User fully connected")

    def list_raw(self):
        # get raw list; used by _list, _query and _query_list
        remote_path = os.path.join(urllib.parse.unquote(self.parsed_url.path.lstrip(u'/')),
                                   self.fakeroot.lstrip(u'/')).rstrip()
        commandline = ((self.cmd + self.auth_switch + u" --auth-list --device-id={0} {1}@{2}::home/{3}"
                       .format(self.idrivedevid, self.idriveid, self.idriveserver, remote_path)))
        try:
            _, l, _ = self.subprocess_popen(commandline)
        except:
            # error: treat as empty response
            log.Debug(u"list EMPTY response ")
            return []

        log.Debug(u"list response: {0}".format(l))

        # get a list of lists from data lines returned by idevsutil_dedup --auth-list
        filtered = map((lambda line: re.split(r"\[|\]", line)), [x for x in l.splitlines() if x.startswith(u"[")])
        # remove whitespace from elements
        filtered = map((lambda line: map((lambda c: c.strip()), line)), filtered)
        # remove empty elements
        filtered = list(map((lambda cols: list(filter((lambda c: c != u''), cols))), filtered))

        return filtered

    def _put(self, source_path, remote_filename):
        # Put a file.
        log.Debug(u"_PUT")
        if not self.user_connected():
            self.connect()

        # decode from byte-stream to utf-8 string
        filename = remote_filename.decode(u'utf-8')

        intrim_file = os.path.join(self.fakeroot, filename)
        remote_dirpath = urllib.parse.unquote(self.parsed_url.path.lstrip(u'/'))

        os.rename(source_path.name, intrim_file)

        log.Debug(u"put_file: source_path={0}, remote_file={1}".format(source_path.name, filename))

        flist = tempfile.NamedTemporaryFile(u'w')
        flist.write(intrim_file)
        flist.seek(0)

        putrequest = ((self.cmd + self.auth_switch + u"  --device-id={0} --files-from={1} / {2}@{3}::home/{4}")
                      .format(self.idrivedevid, flist.name, self.idriveid, self.idriveserver, remote_dirpath))
        log.Debug(u"put_file put command: {0}".format(putrequest))
        _, putresponse, _ = self.subprocess_popen(putrequest)
        log.Debug(u"put_file put response: {0}".format(putresponse))

        flist.close()
        os.remove(intrim_file)

    def _get(self, remote_filename, local_path):
        # Get a file.
        log.Debug(u"_GET")
        if not self.user_connected():
            self.connect()

        # decode from byte-stream to utf-8 string
        filename = remote_filename.decode(u'utf-8')

        remote_path = os.path.join(urllib.parse.unquote(self.parsed_url.path.lstrip(u'/')),
                                   self.fakeroot.lstrip(u'/'), filename).rstrip()

        log.Debug(u"_get: remote_filename={0}, local_path={1}, remote_path={2}, parsed_url.path={3}"
                  .format(filename, local_path, remote_path, self.parsed_url.path))

        # Create tempdir to downlaod file into
        tmpdir = tempfile.mkdtemp()
        log.Debug(u"_get created temporary download folder: {}".format(tmpdir))

        # The filelist file
        flist = tempfile.NamedTemporaryFile(u'w')
        flist.write(remote_path)
        flist.seek(0)

        commandline = ((self.cmd + self.auth_switch + u" --device-id={0} --files-from={1} {2}@{3}::home/ {4}")
                       .format(self.idrivedevid, flist.name, self.idriveid, self.idriveserver, tmpdir))
        log.Debug(u"get command: {0}".format(commandline))
        _, getresponse, _ = self.subprocess_popen(commandline)
        log.Debug(u"_get response: {0}".format(getresponse))

        flist.close()

        # move to the final location
        downloadedSrcPath = os.path.join(tmpdir, remote_path.lstrip(u'/').rstrip(u'/'))
        log.Debug(u"_get moving file {0} to final location: {1}".format(downloadedSrcPath, local_path.name))

        os.rename(downloadedSrcPath, local_path.name)

        shutil.rmtree(tmpdir)

    def _list(self):
        # List files on remote folder
        log.Debug(u"_LIST")
        if not self.user_connected():
            self.connect()

        filtered = self.list_raw()
        filtered = [x[-1] for x in filtered]

        return filtered

    def _delete(self, remote_filename):
        # Delete single file
        log.Debug(u"_DELETE")
        if not self.user_connected():
            self.connect()

        # decode from byte-stream to utf-8 string
        filename = remote_filename.decode(u'utf-8')

        # create a file-list file
        flist = tempfile.NamedTemporaryFile(u'w')
        flist.write(filename.lstrip(u'/'))
        flist.seek(0)

        # target path (remote) on IDrive
        remote_path = os.path.join(urllib.parse.unquote(self.parsed_url.path.lstrip(u'/')),
                                   self.fakeroot.lstrip(u'/')).rstrip()
        log.Debug(u"delete: {0} from remote file path {1}".format(filename, remote_path))

        # delete files from file-list
        delrequest = ((self.cmd + self.auth_switch +
                       u" --delete-items --device-id={0} --files-from={1} {2}@{3}::home/{4}")
                      .format(self.idrivedevid, flist.name, self.idriveid, self.idriveserver, remote_path))
        log.Debug(u"delete: {0}".format(delrequest))
        _, delresponse, _ = self.subprocess_popen(delrequest)
        log.Debug(u"delete response: {0}".format(delresponse))

        # close tempfile
        flist.close()

    def _delete_list(self, filename_list):
        # Delete multiple files
        log.Debug(u"_DELETE LIST")
        if not self.user_connected():
            self.connect()

        # create a file-list file
        flist = tempfile.NamedTemporaryFile(u'w')

        # create file-list
        for filename in filename_list:
            flist.write(filename.decode(u'utf-8').lstrip(u'/') + u'\n')
        flist.seek(0)

        # target path (remote) on IDrive
        remote_path = os.path.join(urllib.parse.unquote(self.parsed_url.path.lstrip(u'/')),
                                   self.fakeroot.lstrip(u'/')).rstrip()
        log.Debug(u"delete multiple files from remote file path {0}".format(remote_path))

        # delete files from file-list
        delrequest = ((self.cmd + self.auth_switch +
                       u" --delete-items --device-id={0} --files-from={1} {2}@{3}::home/{4}")
                      .format(self.idrivedevid, flist.name, self.idriveid, self.idriveserver, remote_path))
        log.Debug(u"delete: {0}".format(delrequest))
        _, delresponse, _ = self.subprocess_popen(delrequest)
        log.Debug(u"delete response: {0}".format(delresponse))

        # close tempfile
        flist.close()

    def _close(self):
        # Remove EVS_temp directory + contents
        log.Debug(u"Removing IDrive temp folder evs_temp")
        try:
            shutil.rmtree(u"evs_temp")
        except:
            pass

    def _query(self, filename):
        log.Debug(u"_QUERY")
        if not self.user_connected():
            self.connect()

        # Get raw directory list; take-out size (index 1) for requested filename (index -1)
        filtered = self.list_raw()
        if filtered:
            filtered = [x[1] for x in filtered if x[-1] == filename.decode(u'utf-8')]
        if filtered:
            return {u'size': int(filtered[0])}

        return {u'size': -1}

    def _query_list(self, filename_list):
        log.Debug(u"_QUERY_LIST")
        if not self.user_connected():
            self.connect()

        # Get raw directory list
        filtered = self.list_raw()

        # For each filename in list: take-out size (index 1) for requested filename (index -1)
        info = {}
        for filename in filename_list:
            if filtered:
                result = [x[1] for x in filtered if x[-1] == filename.decode(u'utf-8')]
            if result:
                info[filename] = {u'size': int(result[0])}
            else:
                info[filename] = {u'size': -1}

        return info

    def __del__(self):
        # remove the self-created temp dir.
        # We do it here, AFTER the clean-up of Duplicity, so it will be empty!
        if self.cleanup:
            os.rmdir(self.fakeroot)


duplicity.backend.register_backend(u"idrived", IDriveBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2020 Jose L. Domingo Lopez <github@24x7linux.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function

import os
import re
import subprocess

import duplicity.backend
from duplicity import util
from duplicity.errors import BackendException
from future import standard_library

standard_library.install_aliases()


class Megav3Backend(duplicity.backend.Backend):
    u"""Backend for MEGA.nz cloud storage, only one that works for accounts created since Nov. 2018
    See https://github.com/megous/megatools/issues/411 for more details

    This MEGA backend resorts to official tools (MEGAcmd) as available at https://mega.nz/cmd
    MEGAcmd works through a single binary called "mega-cmd", which keeps state (for example,
    persisting a session). Multiple "mega-*" shell wrappers (ie. "mega-ls") exist as the user
    interface to "mega-cmd" and MEGA API
    The full MEGAcmd User Guide can be found in the software's GitHub page below :
    https://github.com/meganz/MEGAcmd/blob/master/UserGuide.md"""

    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        # Sanity check : ensure all the necessary "MEGAcmd" binaries exist
        self._check_binary_exists(u'mega-cmd')
        self._check_binary_exists(u'mega-exec')
        self._check_binary_exists(u'mega-help')
        self._check_binary_exists(u'mega-get')
        self._check_binary_exists(u'mega-login')
        self._check_binary_exists(u'mega-logout')
        self._check_binary_exists(u'mega-ls')
        self._check_binary_exists(u'mega-mkdir')
        self._check_binary_exists(u'mega-put')
        self._check_binary_exists(u'mega-rm')
        self._check_binary_exists(u'mega-whoami')

        # "MEGAcmd" does not use a config file, however it is handy to keep one (with the old ".megarc" format) to
        # securely store the username and password
        self._hostname = parsed_url.hostname
        if parsed_url.username is None:
            self._megarc = os.getenv(u'HOME') + u'/.megav3rc'
            try:
                conf_file = open(self._megarc, u"r")
            except Exception as e:
                raise BackendException(
                    u"No password provided in URL and MEGA configuration "
                    u"file for duplicity does not exist as '%s'"
                    % (self._megarc,)
                )

            myvars = {}
            for line in conf_file:
                name, var = line.partition(u"=")[::2]
                myvars[name.strip()] = str(var.strip())
            conf_file.close()
            self._username = myvars[u"Username"]
            self._password = myvars[u"Password"]

        else:
            self._username = parsed_url.username
            self._password = parsed_url.password

        no_logout_option = parsed_url.query_args.get(u'no_logout', [])
        self._no_logout = (len(no_logout_option) > 0) and (
            no_logout_option[0].lower() in [u'1', u'yes', u'true']
        )

        self.ensure_mega_cmd_running()

        # Remote folder ("MEGAcmd" no longer shows "Root/" at the top of the hierarchy)
        self._folder = u'/' + parsed_url.path[1:]

        # Only create the remote folder if it doesn't exist yet
        self.mega_login()
        cmd = [u'mega-ls', self._folder]
        try:
            self.subprocess_popen(cmd)
        except Exception as e:
            self._makedir(self._folder)

    def _check_binary_exists(self, cmd):
        u'Checks that a specified command exists in the running user command path'

        try:
            # Ignore the output, as we only need the return code
            subprocess.check_output([u'which', cmd])
        except Exception as e:
            raise BackendException(
                u"Command '%s' not found, make sure 'MEGAcmd' tools (https://mega.nz/cmd) is "
                u"properly installed and in the running user command path"
                % (cmd,)
            )

    def ensure_mega_cmd_running(self):
        u'Trigger any mega command to ensure mega-cmd server is running'
        try:
            subprocess.run(
                u"mega-help",
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            ).check_returncode()
        except Exception:
            raise BackendException(u'Cannot execute mega command')

    def _makedir(self, path):
        u'Creates a remote directory (recursively if necessary)'

        self.mega_login()
        cmd = [u'mega-mkdir', u'-p', path]
        try:
            self.subprocess_popen(cmd)
        except Exception as e:
            error_str = str(e)
            if u"Folder already exists" in error_str:
                raise BackendException(
                    u"Folder '%s' could not be created on MEGA because it already exists. "
                    u"Use another path or remove the folder in MEGA manually"
                    % (path,)
                )
            else:
                raise BackendException(
                    u"Folder '%s' could not be created, reason : '%s'"
                    % (path, e)
                )

    def _put(self, source_path, remote_filename):
        u"""Uploads file to the specified remote folder (tries to delete it first to make
        sure the new one can be uploaded)"""

        try:
            self.delete(remote_filename.decode())
        except Exception:
            pass
        self.upload(
            local_file=source_path.get_canonical().decode(),
            remote_file=remote_filename.decode(),
        )

    def _get(self, remote_filename, local_path):
        u'Downloads file from the specified remote path'

        self.download(
            remote_file=remote_filename.decode(),
            local_file=local_path.name.decode(),
        )

    def _list(self):
        u'Lists files in the specified remote path'

        return self.folder_contents(files_only=True)

    def _delete(self, filename):
        u'Deletes file from the specified remote path'

        self.delete(remote_file=filename.decode())

    def _close(self):
        u'Function called when backend is done being used'

        if not self._no_logout:
            cmd = [u'mega-logout']
            self.subprocess_popen(cmd)

        cmd = [u'mega-exec', u'exit']
        self.subprocess_popen(cmd)

    def mega_login(self):
        u"""Helper function to check existing session exists"""

        # Abort if command doesn't return in a reasonable time (somehow "mega-session" sometimes
        # doesn't return), and create session if one doesn't exist yet
        try:
            result = subprocess.run(
                u'mega-whoami',
                timeout=30,
                capture_output=True,
            )
            result.check_returncode()
            current_username = result.stdout.decode().split(u':')[-1].strip()
            if current_username != self._username:
                raise Exception(u"Username is not match")
        except subprocess.TimeoutExpired:
            self._close()
            raise BackendException(
                u"Timed out while trying to determine if a MEGA session exists"
            )
        except Exception as e:
            if self._password is None:
                self._password = self.get_password()

            cmd = [u'mega-login', self._username, self._password]
            try:
                subprocess.run(
                    cmd,
                    stderr=subprocess.DEVNULL,
                ).check_returncode()
            except Exception as e:
                self._close()
                raise BackendException(
                    u"Could not log in to MEGA, error : '%s'" % (e,)
                )

    def folder_contents(self, files_only=False):
        u'Lists contents of a remote MEGA path, optionally ignoring subdirectories'

        cmd = [u'mega-ls', u'-l', self._folder]

        self.mega_login()
        files = subprocess.check_output(cmd)
        files = files.decode().split(u'\n')

        # Optionally ignore directories
        if files_only:
            files = [f.split()[5] for f in files if re.search(u'^-', f)]

        return files

    def download(self, remote_file, local_file):
        u'Downloads a file from a remote MEGA path'

        cmd = [u'mega-get', self._folder + u'/' + remote_file, local_file]
        self.mega_login()
        self.subprocess_popen(cmd)

    def upload(self, local_file, remote_file):
        u'Uploads a file to a remote MEGA path'

        cmd = [u'mega-put', local_file, self._folder + u'/' + remote_file]
        self.mega_login()
        try:
            self.subprocess_popen(cmd)
        except Exception as e:
            error_str = str(e)
            if u"Reached storage quota" in error_str:
                raise BackendException(
                    u"MEGA account over quota, could not write file : '%s' . "
                    u"Upgrade your storage at https://mega.nz/pro or remove some data."
                    % (remote_file,)
                )
            else:
                raise BackendException(
                    u"Failed writing file '%s' to MEGA, reason : '%s'"
                    % (remote_file, e)
                )

    def delete(self, remote_file):
        u'Deletes a file from a remote MEGA path'

        cmd = [u'mega-rm', u'-f', self._folder + u'/' + remote_file]
        self.mega_login()
        self.subprocess_popen(cmd)


duplicity.backend.register_backend(u'megav3', Megav3Backend)
duplicity.backend.uses_netloc.extend([u'megav3'])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import map
from builtins import object
from builtins import range

import io
import os
import platform
import unittest

from duplicity import diffdir
from duplicity import patchdir
from duplicity import selection
from duplicity import tarfile
from duplicity import librsync
from duplicity.lazy import *  # pylint: disable=unused-wildcard-import,redefined-builtin
from duplicity.path import *  # pylint: disable=unused-wildcard-import,redefined-builtin
from testing import _runtest_dir
from . import UnitTestCase


class PatchingTest(UnitTestCase):
    u"""Test patching"""
    def setUp(self):
        super(PatchingTest, self).setUp()
        self.unpack_testfiles()

    def copyfileobj(self, infp, outfp):
        u"""Copy in fileobj to out, closing afterwards"""
        blocksize = 32 * 1024
        while 1:
            buf = infp.read(blocksize)
            if not buf:
                break
            outfp.write(buf)
        assert not infp.close()
        assert not outfp.close()

    def test_total(self):
        u"""Test cycle on dirx"""
        self.total_sequence([u'{0}/testfiles/dir1'.format(_runtest_dir),
                             u'{0}/testfiles/dir2'.format(_runtest_dir),
                             u'{0}/testfiles/dir3'.format(_runtest_dir)])

    def get_sel(self, path):
        u"""Get selection iter over the given directory"""
        return selection.Select(path).set_iter()

    def total_sequence(self, filelist):
        u"""Test signatures, diffing, and patching on directory list"""
        assert len(filelist) >= 2
        sig = Path(u"{0}/testfiles/output/sig.tar".format(_runtest_dir))
        diff = Path(u"{0}/testfiles/output/diff.tar".format(_runtest_dir))
        seq_path = Path(u"{0}/testfiles/output/sequence".format(_runtest_dir))
        new_path, old_path = None, None  # set below in for loop

        # Write initial full backup to diff.tar
        for dirname in filelist:
            old_path, new_path = new_path, Path(dirname)
            if old_path:
                sigblock = diffdir.DirSig(self.get_sel(seq_path))
                diffdir.write_block_iter(sigblock, sig)
                deltablock = diffdir.DirDelta(self.get_sel(new_path),
                                              sig.open(u"rb"))
            else:
                deltablock = diffdir.DirFull(self.get_sel(new_path))
            diffdir.write_block_iter(deltablock, diff)

            patchdir.Patch(seq_path, diff.open(u"rb"))
            # print "#########", seq_path, new_path
            assert seq_path.compare_recursive(new_path, 1)

    def test_block_tar(self):
        u"""Test building block tar from a number of files"""
        def get_fileobjs():
            u"""Return iterator yielding open fileobjs of tar files"""
            for i in range(1, 4):
                p = Path(u"{0}/testfiles/blocktartest/test{1}.tar".format(_runtest_dir, i))
                fp = p.open(u"rb")
                yield fp
                fp.close()

        tf = patchdir.TarFile_FromFileobjs(get_fileobjs())
        namelist = []
        for tarinfo in tf:
            namelist.append(tarinfo.name)
        for i in range(1, 6):
            assert (u"tmp/%d" % i) in namelist, namelist

    def test_doubledot_hole(self):
        u"""Test for the .. bug that lets tar overwrite parent dir"""

        def make_bad_tar(filename):
            u"""Write attack tarfile to filename"""
            tf = tarfile.TarFile(name=filename, mode=u"w")

            # file object will be empty, and tarinfo will have path
            # "snapshot/../warning-security-error"
            assert not os.system(u"cat /dev/null > {0}/testfiles/output/file".format(_runtest_dir))
            path = Path(u"{0}/testfiles/output/file".format(_runtest_dir))
            path.index = (b"diff", b"..", b"warning-security-error")
            ti = path.get_tarinfo()
            fp = io.StringIO(u"")
            tf.addfile(ti, fp)

            tf.close()

        make_bad_tar(u"{0}/testfiles/output/bad.tar".format(_runtest_dir))
        os.mkdir(u"{0}/testfiles/output/temp".format(_runtest_dir))

        self.assertRaises(patchdir.PatchDirException, patchdir.Patch,
                          Path(u"{0}/testfiles/output/temp".format(_runtest_dir)),
                          open(u"{0}/testfiles/output/bad.tar".format(_runtest_dir), u"rb"))
        assert not Path(u"{0}/testfiles/output/warning-security-error".format(_runtest_dir)).exists()


class index(object):
    u"""Used below to test the iter collation"""
    def __init__(self, index):
        self.index = index


class CollateItersTest(UnitTestCase):
    def setUp(self):
        super(CollateItersTest, self).setUp()
        self.unpack_testfiles()

    def test_collate(self):
        u"""Test collate_iters function"""
        indicies = [index(i) for i in [0, 1, 2, 3]]
        helper = lambda i: indicies[i]

        makeiter1 = lambda: iter(indicies)
        makeiter2 = lambda: map(helper, [0, 1, 3])
        makeiter3 = lambda: map(helper, [1, 2])

        outiter = patchdir.collate_iters([makeiter1(), makeiter2()])
        assert Iter.equal(outiter,
                          iter([(indicies[0], indicies[0]),
                                (indicies[1], indicies[1]),
                                (indicies[2], None),
                                (indicies[3], indicies[3])]))

        assert Iter.equal(patchdir.collate_iters([makeiter1(),
                                                 makeiter2(),
                                                 makeiter3()]),
                          iter([(indicies[0], indicies[0], None),
                                (indicies[1], indicies[1], indicies[1]),
                                (indicies[2], None, indicies[2]),
                                (indicies[3], indicies[3], None)]), 1)

        assert Iter.equal(patchdir.collate_iters([makeiter1(), iter([])]),
                          map(lambda i: (i, None), indicies))
        assert Iter.equal(map(lambda i: (i, None), indicies),
                          patchdir.collate_iters([makeiter1(), iter([])]))

    def test_tuple(self):
        u"""Test indexed tuple"""
        i = patchdir.IndexedTuple((1, 2, 3), (u"a", u"b"))
        i2 = patchdir.IndexedTuple((), (u"hello", u"there", u"how are you"))

        assert i[0] == u"a"
        assert i[1] == u"b"
        assert i2[1] == u"there"
        assert len(i) == 2 and len(i2) == 3
        assert i2 < i, i2 < i

    def test_tuple_assignment(self):
        a, b, c = patchdir.IndexedTuple((), (1, 2, 3))
        assert a == 1
        assert b == 2
        assert c == 3


class TestInnerFuncs(UnitTestCase):
    u"""Test some other functions involved in patching"""
    def setUp(self):
        super(TestInnerFuncs, self).setUp()
        self.unpack_testfiles()
        self.check_output()

    def check_output(self):
        u"""Make {0}/testfiles/output exists"""
        out = Path(u"{0}/testfiles/output".format(_runtest_dir))
        if not (out.exists() and out.isdir()):
            out.mkdir()
        self.out = out

    def snapshot(self):
        u"""Make a snapshot ROPath, permissions 0o600"""
        ss = self.out.append(u"snapshot")
        fout = ss.open(u"wb")
        fout.write(b"hello, world!")
        assert not fout.close()
        ss.chmod(0o600)
        ss.difftype = u"snapshot"
        return ss

    def get_delta(self, old_buf, new_buf):
        u"""Return delta buffer from old to new"""
        sigfile = librsync.SigFile(io.BytesIO(old_buf))
        sig = sigfile.read()
        assert not sigfile.close()

        deltafile = librsync.DeltaFile(sig, io.BytesIO(new_buf))
        deltabuf = deltafile.read()
        assert not deltafile.close()
        return deltabuf

    def delta1(self):
        u"""Make a delta ROPath, permissions 0o640"""
        delta1 = self.out.append(u"delta1")
        fout = delta1.open(u"wb")
        fout.write(self.get_delta(b"hello, world!",
                                  b"aonseuth aosetnuhaonsuhtansoetuhaoe"))
        assert not fout.close()
        delta1.chmod(0o640)
        delta1.difftype = u"diff"
        return delta1

    def delta2(self):
        u"""Make another delta ROPath, permissions 0o644"""
        delta2 = self.out.append(u"delta1")
        fout = delta2.open(u"wb")
        fout.write(self.get_delta(b"aonseuth aosetnuhaonsuhtansoetuhaoe",
                                  b"3499 34957839485792357 458348573"))
        assert not fout.close()
        delta2.chmod(0o644)
        delta2.difftype = u"diff"
        return delta2

    def deleted(self):
        u"""Make a deleted ROPath"""
        deleted = self.out.append(u"deleted")
        assert not deleted.exists()
        deleted.difftype = u"deleted"
        return deleted

    def test_normalize(self):
        u"""Test normalizing a sequence of diffs"""
        ss = self.snapshot()
        d1 = self.delta1()
        d2 = self.delta2()
        de = self.deleted()

        seq1 = [ss, d1, d2]
        seq2 = [ss, d1, d2, de]
        seq3 = [de, ss, d1, d2]
        seq4 = [de, ss, d1, d2, ss]
        seq5 = [de, ss, d1, d2, ss, d1, d2]

        def try_seq(input_seq, correct_output_seq):
            normed = patchdir.normalize_ps(input_seq)
            assert normed == correct_output_seq, (normed, correct_output_seq)

        try_seq(seq1, seq1)
        try_seq(seq2, [de])
        try_seq(seq3, seq1)
        try_seq(seq4, [ss])
        try_seq(seq5, seq1)

    # TODO: fix test_patch_seq2ropath for macOS, maybe others.
    #       Fails under tox, pytest, and pydevd
    #----------
    #     def testseq(seq, perms, buf):
    #         result = patchdir.patch_seq2ropath(seq)
    # >       assert result.getperms() == perms, (result.getperms(), perms)
    # E       AssertionError: ('501:0 600', '501:20 600')
    # E       assert '501:0 600' == '501:20 600'
    # E         - 501:0 600
    # E         + 501:20 600
    @unittest.skipUnless(platform.platform().startswith(u"Linux"), u"Skip on non-Linux systems")
    def test_patch_seq2ropath(self):
        u"""Test patching sequence"""
        def testseq(seq, perms, buf):
            result = patchdir.patch_seq2ropath(seq)
            assert result.getperms() == perms, (result.getperms(), perms)
            fout = result.open(u"rb")
            contents = fout.read()
            assert not fout.close()
            assert contents == buf, (contents, buf)

        ids = u"%d:%d" % (os.getuid(), os.getgid())

        testseq([self.snapshot()], (u"%s 600" % ids), b"hello, world!")
        testseq([self.snapshot(), self.delta1()], (u"%s 640" % ids),
                b"aonseuth aosetnuhaonsuhtansoetuhaoe")
        testseq([self.snapshot(), self.delta1(), self.delta2()], (u"%s 644" % ids),
                b"3499 34957839485792357 458348573")


if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2014 Michael Terry <michael.terry@canonical.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import os
import sys
import subprocess
import pytest
import fnmatch
import os

if os.getenv(u'RUN_CODE_TESTS', None) == u'1':
    # Make conditional so that we do not have to import in environments that
    # do not run the tests (e.g. the build servers)
    import pycodestyle

from . import _top_dir, DuplicityTestCase
from . import find_unadorned_strings

skipCodeTest = pytest.mark.skipif(not os.getenv(u'RUN_CODE_TESTS', None) == u'1',
                                  reason=u'Must set environment var RUN_CODE_TESTS=1')


class CodeTest(DuplicityTestCase):

    def run_checker(self, cmd, returncodes=[0]):
        process = subprocess.Popen(cmd,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   universal_newlines=True)
        output = process.communicate()[0]
        if len(output):
            for line in output.split(u'\n'):
                print(line, file=sys.stderr)
            output = u""
        self.assertTrue(process.returncode in returncodes, output)

    @skipCodeTest
    def test_2to3(self):
        # As we modernize the source code, we can remove more and more nofixes
        self.run_checker([
            u"2to3",
            u"--nofix=next",
            u"--nofix=types",
            u"--nofix=unicode",
            # The following fixes we don't want to remove, since they are false
            # positives, things we don't care about, or real incompatibilities
            # but which 2to3 can fix for us better automatically.
            u"--nofix=callable",
            u"--nofix=dict",
            u"--nofix=future",
            u"--nofix=imports",
            u"--nofix=print",
            u"--nofix=raw_input",
            u"--nofix=urllib",
            u"--nofix=xrange",
            u"--nofix=map",
            _top_dir])

    @skipCodeTest
    def test_pylint(self):
        u"""Pylint test (requires pylint to be installed to pass)"""
        self.run_checker([
            u"pylint",
            u"--rcfile=" + os.path.join(_top_dir, u"pylintrc"),
            os.path.join(_top_dir, u'duplicity'),
            os.path.join(_top_dir, u'bin/duplicity'),
            os.path.join(_top_dir, u'bin/rdiffdir')])

    @skipCodeTest
    def test_pep8(self):
        u"""Test that we conform to PEP-8 using pycodestyle."""
        # Note that the settings, ignores etc for pycodestyle are set in tox.ini, not here
        style = pycodestyle.StyleGuide(config_file=os.path.join(_top_dir, u'tox.ini'))
        result = style.check_files([os.path.join(_top_dir, u'duplicity'),
                                    os.path.join(_top_dir, u'bin/duplicity'),
                                    os.path.join(_top_dir, u'bin/rdiffdir')])
        self.assertEqual(result.total_errors, 0,
                         u"Found %s code style errors (and warnings)." % result.total_errors)

    @skipCodeTest
    def test_unadorned_string_literals(self):
        u"""For predictable results in python/3 all string literals need to be marked as unicode, bytes or raw"""

        ignored_files = [
                         # These are not source files we want to check
                         os.path.join(_top_dir, u'.tox', u'*'),
                         os.path.join(_top_dir, u'.eggs', u'*'),
                         os.path.join(_top_dir, u'docs', u'conf.py'),
                         # TODO Every file from here down needs to be fixed and the exclusion removed
                         ]


        # Find all the .py files in the duplicity tree
        # We cannot use glob.glob recursive until we drop support for Python < 3.5
        matches = []

        def multi_filter(names, patterns):
            u"""Generator function which yields the names that match one or more of the patterns."""
            for name in names:
                if any(fnmatch.fnmatch(name, pattern) for pattern in patterns):
                    yield name

        for root, dirnames, filenames in os.walk(_top_dir):
            for filename in fnmatch.filter(filenames, u'*.py'):
                matches.append(os.path.join(root, filename))

        excluded = multi_filter(matches, ignored_files) if ignored_files else []
        matches = list(set(matches) - set(excluded))
        matches.extend([
            os.path.join(_top_dir, u"bin", u"duplicity"),
            os.path.join(_top_dir, u"bin", u"rdiffdir")
            ])

        do_assert = False
        for python_source_file in matches:
            # Check each of the relevant python sources for unadorned string literals
            unadorned_string_list = find_unadorned_strings.check_file_for_unadorned(python_source_file)
            if unadorned_string_list:
                do_assert = True
                print(u"Found {0:d} unadorned strings in {1:s}:".format(
                    len(unadorned_string_list), python_source_file),
                    file=sys.stderr)
                for unadorned_string in unadorned_string_list:
                    print(unadorned_string[1:], file=sys.stderr)

        self.assertEqual(do_assert, False, u"Found unadorned strings.")


if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2011 Henrique Carvalho Alves <hcarvalhoalves@gmail.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import duplicity.backend
from duplicity import config

if config.s3_use_multiprocessing:
    from ._boto_multi import BotoBackend
else:
    from ._boto_single import BotoBackend

duplicity.backend.register_backend(u"gs", BotoBackend)
duplicity.backend.register_backend(u"s3", BotoBackend)
duplicity.backend.register_backend(u"s3+http", BotoBackend)
duplicity.backend.uses_netloc.extend([u's3'])
# s3 is also implemented by the newer boto3 backend now
duplicity.backend.register_backend(u"boto+s3", BotoBackend)
duplicity.backend.uses_netloc.extend([u'boto+s3'])



from __future__ import print_function
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import pytest
import random
import unittest

from . import UnitTestCase
from duplicity import backend
from duplicity import config
from duplicity import dup_collections
from duplicity import dup_time
from duplicity import gpg
from duplicity import path
from testing import _runtest_dir

filename_list1 = [b"duplicity-full.2002-08-17T16:17:01-07:00.manifest.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol1.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol2.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol3.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol4.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol5.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol6.difftar.gpg",
                  b"duplicity-inc.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.manifest.gpg",
                  b"duplicity-inc.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.vol1.difftar.gpg",
                  b"Extra stuff to be ignored"]

remote_sigchain_filename_list = [b"duplicity-full-signatures.2002-08-17T16:17:01-07:00.sigtar.gpg",
                                 b"duplicity-new-signatures.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.sigtar.gpg",
                                 b"duplicity-new-signatures.2002-08-18T00:04:30-07:00.to.2002-08-20T00:00:00-07:00.sigtar.gpg"]

local_sigchain_filename_list = [b"duplicity-full-signatures.2002-08-17T16:17:01-07:00.sigtar.gz",
                                b"duplicity-new-signatures.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.sigtar.gz",
                                b"duplicity-new-signatures.2002-08-18T00:04:30-07:00.to.2002-08-20T00:00:00-07:00.sigtar.gz"]

# A filename list with some incomplete volumes, an older full volume,
# and a complete chain.
filename_list2 = [b"duplicity-full.2001-01-01T16:17:01-07:00.manifest.gpg",
                  b"duplicity-full.2001-01-01T16:17:01-07:00.vol1.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.manifest.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol1.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol2.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol3.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol4.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol5.difftar.gpg",
                  b"duplicity-full.2002-08-17T16:17:01-07:00.vol6.difftar.gpg",
                  b"duplicity-inc.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.manifest.gpg",
                  b"duplicity-inc.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.vol1.difftar.gpg",
                  b"The following are extraneous duplicity files",
                  b"duplicity-new-signatures.2001-08-17T02:05:13-05:00.to.2002-08-17T05:05:14-05:00.sigtar.gpg",
                  b"duplicity-full.2002-08-15T01:01:01-07:00.vol1.difftar.gpg",
                  b"duplicity-inc.2000-08-17T16:17:01-07:00.to.2000-08-18T00:04:30-07:00.manifest.gpg",
                  b"duplicity-inc.2000-08-17T16:17:01-07:00.to.2000-08-18T00:04:30-07:00.vol1.difftar.gpg",
                  b"Extra stuff to be ignored"]


class CollectionTest(UnitTestCase):
    u"""Test collections"""
    def setUp(self):
        super(CollectionTest, self).setUp()

        self.unpack_testfiles()

        col_test_dir = path.Path(u"{0}/testfiles/collectionstest".format(_runtest_dir))
        archive_dir_path = col_test_dir.append(u"archive_dir")
        self.set_config(u'archive_dir_path', archive_dir_path)
        self.archive_dir_backend = backend.get_backend(u"file://{0}/testfiles/collectionstest".format(_runtest_dir) +
                                                       u"/archive_dir")

        self.real_backend = backend.get_backend(u"file://%s/%s" %
                                                (col_test_dir.uc_name, u"remote_dir"))
        self.output_dir = path.Path(u"{0}/testfiles/output".format(_runtest_dir))  # used as a temp directory
        self.output_dir_backend = backend.get_backend(u"file://{0}/testfiles/output".format(_runtest_dir))

    def set_gpg_profile(self):
        u"""Set gpg profile to standard "foobar" sym"""
        self.set_config(u'gpg_profile', gpg.GPGProfile(passphrase=u"foobar"))

    def test_backup_chains(self):
        u"""Test basic backup chain construction"""
        random.shuffle(filename_list1)
        cs = dup_collections.CollectionsStatus(None, config.archive_dir_path, u"full")
        chains, orphaned, incomplete = cs.get_backup_chains(filename_list1)
        if len(chains) != 1 or len(orphaned) != 0:
            print(chains)
            print(orphaned)
            assert 0

        chain = chains[0]
        assert chain.end_time == 1029654270
        assert chain.fullset.time == 1029626221

    def test_collections_status(self):
        u"""Test CollectionStatus object's set_values()"""
        def check_cs(cs):
            u"""Check values of collections status"""
            assert cs.values_set

            assert cs.matched_chain_pair
            assert cs.matched_chain_pair[0].end_time == 1029826800
            assert len(cs.all_backup_chains) == 1, cs.all_backup_chains

        cs = dup_collections.CollectionsStatus(self.real_backend, config.archive_dir_path, u"full").set_values()
        check_cs(cs)
        assert cs.matched_chain_pair[0].islocal()

    def test_sig_chain(self):
        u"""Test a single signature chain"""
        chain = dup_collections.SignatureChain(1, config.archive_dir_path)
        for filename in local_sigchain_filename_list:
            assert chain.add_filename(filename)
        assert not chain.add_filename(b"duplicity-new-signatures.2002-08-18T00:04:30-07:00.to.2002-08-20T00:00:00-07:00.sigtar.gpg")

    def test_sig_chains(self):
        u"""Test making signature chains from filename list"""
        cs = dup_collections.CollectionsStatus(None, config.archive_dir_path, u"full")
        chains, orphaned_paths = cs.get_signature_chains(local=1)
        self.sig_chains_helper(chains, orphaned_paths)

    def test_sig_chains2(self):
        u"""Test making signature chains from filename list on backend"""
        cs = dup_collections.CollectionsStatus(self.archive_dir_backend, config.archive_dir_path, u"full")
        chains, orphaned_paths = cs.get_signature_chains(local=None)
        self.sig_chains_helper(chains, orphaned_paths)

    def sig_chains_helper(self, chains, orphaned_paths):
        u"""Test chains and orphaned_paths values for two above tests"""
        if orphaned_paths:
            for op in orphaned_paths:
                print(op)
            assert 0
        assert len(chains) == 1, chains
        assert chains[0].end_time == 1029826800

    def sigchain_fileobj_get(self, local):
        u"""Return chain, local if local is true with filenames added"""
        if local:
            chain = dup_collections.SignatureChain(1, config.archive_dir_path)
            for filename in local_sigchain_filename_list:
                assert chain.add_filename(filename)
        else:
            chain = dup_collections.SignatureChain(None, self.real_backend)
            for filename in remote_sigchain_filename_list:
                assert chain.add_filename(filename)
        return chain

    def sigchain_fileobj_check_list(self, chain):
        u"""Make sure the list of file objects in chain has right contents

        The contents of the /tmp/testfiles/collectiontest/remote_dir have
        to be coordinated with this test.

        """
        fileobjlist = chain.get_fileobjs()
        assert len(fileobjlist) == 3

        def test_fileobj(i, s):
            buf = fileobjlist[i].read()
            fileobjlist[i].close()
            assert buf == s, (buf, s)

        test_fileobj(0, b"Hello, world!")
        test_fileobj(1, b"hello 1")
        test_fileobj(2, b"Hello 2")

    @pytest.mark.usefixtures(u"redirect_stdin")
    def test_sigchain_fileobj(self):
        u"""Test getting signature chain fileobjs from archive_dir_path"""
        self.set_gpg_profile()
        self.sigchain_fileobj_check_list(self.sigchain_fileobj_get(1))
        self.sigchain_fileobj_check_list(self.sigchain_fileobj_get(None))

    def get_filelist2_cs(self):
        u"""Return set CollectionsStatus object from filelist 2"""
        # Set up /tmp/testfiles/output with files from filename_list2
        for filename in filename_list2:
            p = self.output_dir.append(filename)
            p.touch()

        cs = dup_collections.CollectionsStatus(self.output_dir_backend, config.archive_dir_path, u"full")
        cs.set_values()
        return cs

    def test_get_extraneous(self):
        u"""Test the listing of extraneous files"""
        cs = self.get_filelist2_cs()
        assert len(cs.orphaned_backup_sets) == 1, cs.orphaned_backup_sets
        assert len(cs.local_orphaned_sig_names) == 0, cs.local_orphaned_sig_names
        assert len(cs.remote_orphaned_sig_names) == 1, cs.remote_orphaned_sig_names
        assert len(cs.incomplete_backup_sets) == 1, cs.incomplete_backup_sets

        right_list = [b"duplicity-new-signatures.2001-08-17T02:05:13-05:00.to.2002-08-17T05:05:14-05:00.sigtar.gpg",
                      b"duplicity-full.2002-08-15T01:01:01-07:00.vol1.difftar.gpg",
                      b"duplicity-inc.2000-08-17T16:17:01-07:00.to.2000-08-18T00:04:30-07:00.manifest.gpg",
                      b"duplicity-inc.2000-08-17T16:17:01-07:00.to.2000-08-18T00:04:30-07:00.vol1.difftar.gpg"]
        local_received_list, remote_received_list = cs.get_extraneous()
        errors = []
        for filename in remote_received_list:
            if filename not in right_list:
                errors.append(u"### Got bad extraneous filename " + filename.decode())
            else:
                right_list.remove(filename)
        for filename in right_list:
            errors.append(u"### Didn't receive extraneous filename " + filename)
        assert not errors, u"\n" + u"\n".join(errors)

    def test_get_olderthan(self):
        u"""Test getting list of files older than a certain time"""
        cs = self.get_filelist2_cs()
        oldsets = cs.get_older_than(
            dup_time.genstrtotime(u"2002-05-01T16:17:01-07:00"))
        oldset_times = [s.get_time() for s in oldsets]
        right_times = [dup_time.genstrtotime(u'2001-01-01T16:17:01-07:00')]
        assert oldset_times == right_times, \
            [oldset_times, right_times]

        oldsets_required = cs.get_older_than_required(
            dup_time.genstrtotime(u"2002-08-17T20:00:00-07:00"))
        oldset_times = [s.get_time() for s in oldsets_required]
        right_times_required = [dup_time.genstrtotime(u'2002-08-17T16:17:01-07:00')]
        assert oldset_times == right_times_required, \
            [oldset_times, right_times_required]


if __name__ == u"__main__":
    unittest.main()
